/* compare.c
 * contains the wrapper that accesses comparisons
 * as well as the compare logic for inside the math functions.
 *
 * (c) Copyright 2000 Aaron Kimball
 */


long EvaluateComparison(char *str,FILE *fpin,FILE* fpout,LPSAFELISTITEM pThisLine)
{
	long int x=0,err=0;
	double f=0;

	err = eval_expression(str,&x,&x,&f,6,-1);
	print_eval_error(err);
	if(err == 10) return 10;
	if(err == 11) return 1; //a string is a true value. always.
	if(err) return 0;

	if(f && !x) x=(long)f;
	return x;
}

long DoStringCompare(char *str,long *err)
{
	SAFESTRING left,right;
	char op[4];
	char *p1=0;
	int res=0;
	int CompResult=0;
	SAFESTRING resultstr;
	char *p2=0,*p3=0,*p4;
	int numquotes;

	//ok, now we tokenate on left/op/right
	p4 = str;
find_str_comp_op:
	p1 = strstr(p4,"eq");
	if(!p1) p1 = strstr(p4,"is");
	if(!p1) p1 = strstr(p4,"ne");
	if(!p1)
	{
		*err = 12; //no comparator found
		return 0;
	}

        numquotes = 0;
        for(p2 = str;p2<p1;p2++)
        {
                if(*p2 == '\"')
                {
                        for(p3=p2-1;p3>=str;p3--)
                                if(*p3 != '\\') break;
                        if((p3-p2)%2) numquotes++;
                }
        }
        if(numquotes%2)
	{
		p4 = p1 + 1;
		goto find_str_comp_op;
	}

	if( *(p1-1) != ' ' && *(p1-1) != '\t' && *(p1+2) != ' ' && *(p1+2) != '\t') 
	{
		p4 = p1+1;
		goto find_str_comp_op;
	}


	zerostring(&left);
	zerostring(&right);

	//get the left and right sides...
	addrawstrn(&left,str,p1-str);
	addrawstr(&right,p1+3);
	memset(op,0,4);
	memcpy(op,p1,2);

	//ok, evaluate each side!

	//shake it to the left (if you're having a good time)
	
	zerostring(&resultstr);
	res = do_string_eval_work(&left,&resultstr);
	//(use a temp variable to receive the result. then dump that back into left.str)
	zerostring(&left);
	addsafestr(&left,&resultstr);
	killstring(&resultstr);

	if(res)
	{
		killstring(&left);
		killstring(&right);
		*err = res;
		return 0;
	}

	//same for the right side!

	zerostring(&resultstr);
	res = do_string_eval_work(&right,&resultstr);
	//(use a temp variable to receive the result. then dump that back into left.str)
	zerostring(&right);
	addsafestr(&right,&resultstr);
	killstring(&resultstr);

        if(res)
        {
                killstring(&left);
                killstring(&right);
		*err = res;
                return 0;
        }

	//now that we have the values for each side, lets compare them
	CompResult = 0;
	if(strcmp(op,"eq")==0 || strcmp(op,"is")==0) {
		if(strcmp(left.str,right.str)==0) { CompResult =  1; }
		else { CompResult = 0; }
	}
	if(strcmp(op,"ne")==0) {
		if(strcmp(left.str,right.str) != 0) { CompResult =  1; }
		else { CompResult = 0; }
	}

	killstring(&left);
	killstring(&right);	

//printf("StrComp returns: <%i>\n",CompResult);

	return CompResult;
}

long DoCompareWork(char *str,long *err)
{
	SAFESTRING left,right;
	char op[4];
	
	char *p1=0;
	int res=0;
	long longret=0,L_long=0,R_long=0;
	double dblret=0,L_dbl=0,R_dbl=0;
	int CompResult=0;

//printf("Comp: <%s>\n",str);

	//ok, now we tokenate on left/op/right
	p1 = strstr(str,"eq");
	while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
		{ p1++; if(*p1) p1 = strstr(p1,"eq"); }

	if(!p1)
	{
		p1 = strstr(str,"is");
		while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
			{ p1++; if(*p1) p1 = strstr(p1,"is"); }
	}
	if(!p1)
	{
		p1 = strstr(str,"ne");
		while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
			{ p1++; if(*p1) p1 = strstr(p1,"ne"); }
	}
	if(!p1)
	{
		p1 = strstr(str,"ge");
		while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
			{ p1++; if(*p1) p1 = strstr(p1,"ge"); }
	}
	if(!p1)
	{
		p1 = strstr(str,"gt");
		while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
			{ p1++; if(*p1) p1 = strstr(p1,"gt"); }
	}
	if(!p1)
	{
		p1 = strstr(str,"le");
		while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
			{ p1++; if(*p1) p1 = strstr(p1,"le"); }
	}
	if(!p1)
	{
		p1 = strstr(str,"lt");
		while(p1 && ((*(p1-1) != ' ' && *(p1-1) != '\t') || (*(p1+2) != ' ' && *(p1+2) != '\t')))
			{ p1++; if(*p1) p1 = strstr(p1,"lt"); }
	}
	if(!p1)
	{
		*err = 12; //no comparator found
		return 0;
	}

	zerostring(&left);
	zerostring(&right);

	//get the left and right sides...
	addrawstrn(&left,str,p1-str);
	addrawstr(&right,p1+3);
	memset(op,0,4);
	memcpy(op,p1,2);

	//ok, evaluate each side!

	//shake it to the left (if you're having a good time)
	longret = 0;
	dblret = 0;
	
        res = eval_expression(left.str,&longret,&longret,&dblret,12,0);

	if(res)
	{
		killstring(&left);
		killstring(&right);
		*err = res;
		return 0;
	}

	L_long = longret;
	L_dbl = dblret;

	//shake it to the right (if you're havin a great time)
        longret = 0;
        dblret = 0;

        res = eval_expression(right.str,&longret,&longret,&dblret,12,0);

        if(res)
        {
                killstring(&left);
                killstring(&right);
		*err = res;
                return 0;
        }

        R_long = longret;
        R_dbl = dblret;

	//now that we have the values for each side, lets compare them
	CompResult = 0;
	if(strcmp(op,"eq")==0 || strcmp(op,"is")==0) {
		if((R_dbl == 0 && R_long != 0) || (L_dbl == 0 && L_long != 0))
			{ if(L_long == R_long) CompResult = 1; }
		else { if(L_dbl == R_dbl) CompResult = 1; }
	}
	if(strcmp(op,"ne")==0) {
		if((R_dbl == 0 && R_long != 0) || (L_dbl == 0 && L_long != 0))
			{ if(L_long != R_long) CompResult = 1; }
		else { if(L_dbl != R_dbl) CompResult = 1; }
	}
	if(strcmp(op,"ge")==0) {
		if((R_dbl == 0 && R_long != 0) || (L_dbl == 0 && L_long != 0))
			{ if(L_long >= R_long) CompResult = 1; }
		else { if(L_dbl >= R_dbl) CompResult = 1; }
	}
	if(strcmp(op,"gt")==0) {
		if((R_dbl == 0 && R_long != 0) || (L_dbl == 0 && L_long != 0))
			{ if(L_long > R_long) CompResult = 1; }
		else { if(L_dbl > R_dbl) CompResult = 1; }
	}
	if(strcmp(op,"le")==0) {
		if((R_dbl == 0 && R_long != 0) || (L_dbl == 0 && L_long != 0))
			{ if(L_long <= R_long) CompResult = 1; }
		else { if(L_dbl <= R_dbl) CompResult = 1; }
	}
	if(strcmp(op,"lt")==0) {
		if((R_dbl == 0 && R_long != 0) || (L_dbl == 0 && L_long != 0))
			{ if(L_long < R_long) CompResult = 1; }
		else { if(L_dbl < R_dbl) CompResult = 1; }
	}

	killstring(&left);
	killstring(&right);	

	return CompResult;
}
