//dlls.c
//
//contains the linked-list code
//which controls the dll list.
//
// (c) 2001 Aaron Kimball

LPDLLHANDLELISTITEM pFirstDllHandle;
LPDLLHANDLELISTITEM pLastDllHandle;

#include <dlfcn.h>

int kill_all_dlls()
{
	LPDLLHANDLELISTITEM pNext;
	LPDLLHANDLELISTITEM pThis;

	pNext = pFirstDllHandle;

	while(pNext)
	{
		pThis=pNext;
		pNext=pThis->next;
		dlclose(pThis->handle);
		free(pThis);
	}
	
	return 0;
}

int create_dll_entry(char *filename)
{

	LPDLLHANDLELISTITEM pNew;
	LIBRARYHANDLES hnds;
	int (*reglib)(LPLIBRARYHANDLES);
	char *err;

	pNew = malloc(sizeof(DLLHANDLELISTITEM));

	if(!pNew) return 1; //enomem

	//open it up
	pNew->handle = dlopen(filename,RTLD_LAZY);
	if(pNew->handle == 0)
	{
		dlerror();
		return 2; //file not found.
	}

	//whip up a LIBRARYHANDLES to pass in

	hnds.ver = 1;

	hnds.pFirstFunc = pFirstFunc;
	hnds.pLastFunc = pLastFunc;
	hnds.pFirstLabel = pFirstLabel;
	hnds.pLastLabel = pLastLabel;
	hnds.pFirstItem = pFirstItem;
	hnds.pLastItem = pLastItem;
	hnds.global_cur_line = &global_cur_line;
	hnds.global_cur_user_line = &global_cur_user_line;
	hnds.global_double_return = &global_double_return;
	hnds.pFirstFuncArg = pFirstFuncArg;
	hnds.pLastFuncArg = pLastFuncArg;
	hnds.sub_line_goto = &sub_line_goto;
	hnds.calls = &calls;
	hnds.CallDepth = &CallDepth;

	hnds.strhash = strhash;
	hnds.stripwhite = strip_whitespace;
	hnds.getstrptr = getstrptr;
	hnds.zerostring = zerostring;
	hnds.addrawstr = addrawstr;
	hnds.addrawstrn = addrawstrn;
	hnds.addsafestr = addsafestr;
	hnds.addsafestrn = addsafestrn;
	hnds.addsafestro = addsafestro;
	hnds.addsafestrno = addsafestrno;
	hnds.killstr = killstring;
	hnds.replacestr = replacestr;
	hnds.ltrim = ltrim;
	hnds.ltoa = ltoa;
	hnds.ftoa = ftoa;
	hnds.create_intrinsic = create_intrinsic_function;
	hnds.__emu_create_user_variables = create_user_variables;
	hnds.__emu_eval_expression = eval_expression;
	hnds.__emu_internal_GetUserIdentifier = GetUserIdentifier;
	hnds.__emu_pop = pop;
	hnds.__emu_push = push;
	hnds.__emu_kill_arg_list_item = kill_arg_list_item;
	hnds.__emu_create_arg_list_item = create_arg_list_item;
	hnds.__emu_get_arg_by_num = get_arg_by_num;
	hnds.__emu_get_next_arg = get_next_arg;
	hnds.__emu_get_arg_count = get_arg_count;
	hnds.__emu_get_arg_string = get_arg_string;
	hnds.__emu_add_arg_string = add_arg_string;
	hnds.__emu_pop_arg_string = pop_arg_string;
	hnds.__emu_print_eval_error = print_eval_error;
	hnds.__emu_print_arg_error = print_arg_error;



	//call the RegisterLibrary() to add it into our 
	//list'o'funcs
	dlerror(); //clear the error cache
	reglib = dlsym(pNew->handle,"RegisterLibrary");
	err = dlerror();
	if(err)
	{
		printf("Error linking shared object: %s\n",err);
		return 3; //Register Library not found
	}

	(*reglib)(&hnds);

	pNew->next = 0;
	pNew->prev = pLastDllHandle;	
	if(pLastDllHandle) pLastDllHandle->next = pNew;
	pLastDllHandle = pNew;



	return 0;
}
