/* emu.h
 * EMU interpretter
 * 
 * Some basic defines...
 *
 * (c) Copyright 2000 Aaron Kimball
 *
 */

#ifndef __EMU_STANDARD_INCLUDE
#define __EMU_STANDARD_INCLUDE

#include<stdio.h>
#include<string.h>
#include<memory.h>
#include<ctype.h>
#include<stdlib.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
 
#ifdef TARGET_LINUX
#include<dlfcn.h>
#endif

#ifdef TARGET_DOS
#include<conio.h>
#endif


typedef struct 
{
        char *str;
        long len; //malloc'd length.
	long actual_len; //string length.
} SAFESTRING;

typedef SAFESTRING* LPSAFESTRING;

#define STACK_MAX 32768
typedef struct __num_stack
{
        long data[STACK_MAX];
        int n;
} STACK;

extern STACK calls;

typedef struct __safe_str_list
{
	char *str;
	long len;
	long line_no;
	long sub_line_no;
	long file_no;
        long user_line_no;
	int type; //1 for code, 0 for text, 2 for computer-generated code.
	struct __safe_str_list *next;
        struct __safe_str_list *prev;
} SAFELISTITEM;
typedef SAFELISTITEM* LPSAFELISTITEM;

extern SAFELISTITEM *pFirstItem, *pLastItem;

#define LangFunc(x) int x(char *str,FILE *fpin, FILE *fpout,LPSAFELISTITEM pThisLine,unsigned long *force_jump)

typedef struct __func_def_list
{
        unsigned long hash;
        int type; //0  for intrinsic,
                  //1  for userdefined func,
                  //2  for float,
                  //3  for int,
                  //4  for uint
                  //5  for a string.
                  //6  is internal int,
                  //7  is internal float
                  //8  is float array
                  //9  is int array
                  //10 is uint array
                  //11 is string array.
		  //12 is internal non-comparative float
		  //13 is for intrinsic keyword.
		  //14-17 are reserved for defining constant variables, but degenerate into 2-5.
        union
        {
                unsigned long line; //what line is it on?
                int (*func)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long *); //the func itself.
        } ptr;
        union
        {
                double float_val;
                long long_val;
                unsigned long ulong_val;
                LPSAFESTRING str_val;
        } val;
        int children;
        int domain; //local only? (0 = global)
	char name[41];
        struct __func_def_list *next;
        struct __func_def_list *parent; //array items point to the array.
} FUNCTIONDEF;
typedef FUNCTIONDEF* LPFUNCTIONDEF;

extern FUNCTIONDEF *pFirstFunc, *pLastFunc;

typedef struct __dll_handle_list
{
	void *handle;
	struct __dll_handle_list *next;
	struct __dll_handle_list *prev;
} DLLHANDLELISTITEM;
typedef DLLHANDLELISTITEM* LPDLLHANDLELISTITEM;

extern LPDLLHANDLELISTITEM pFirstDllHandle;
extern LPDLLHANDLELISTITEM pLastDllHandle;


typedef struct __label_list_struct
{
	unsigned long int hash;
	unsigned long int line;
	unsigned long int subline;
	unsigned long int depth;	//function call nest depth
	unsigned long int if_depth;	//how many if-blocks it's inside.
	SAFESTRING name;
	struct __label_list_struct *pPrev;
	struct __label_list_struct *pNext;
} LABELLISTITEM;
typedef LABELLISTITEM* LPLABELLISTITEM;

extern LPLABELLISTITEM pFirstLabel;
extern LPLABELLISTITEM pLastLabel;

extern FILE *fpout_global;
extern int bAbort;
extern int bCopying;
extern int bInFunc;
extern int bInScript;
extern int bSeeking;
extern int seekline;
extern int userseekline;
extern int sub_line_goto;
extern int CallDepth;
extern int CurIfEvaluation;
extern int IfDepth;
extern int LoopsEncountered;
extern unsigned long int labelseekhash;
extern unsigned long int labeldepth;
extern int safemode;
extern int ExecFunc;
extern int cur_file_num;
extern long int global_cur_line;
extern long int global_cur_subline;
extern long int global_cur_user_line;
extern long next_statement;
extern long next_to_exec;
extern long force_jump;
extern double global_double_return;
extern long next_user_line_number;
extern int backgroundmode;
extern int quietenv;

extern SAFESTRING labelseekname;
extern SAFESTRING seekstr;
extern SAFESTRING LoopIncAssign;

typedef struct __file_state
{
        int bCopying;
        FILE *fp;
        int file_num;
        int line_num; //real user line num, not internal line.
        struct __file_state *pNextState;
        struct __file_state *pPrevState;
} FILESTATE;
typedef FILESTATE* LPFILESTATE;

typedef struct __func_args
{
        LPSAFESTRING argstr;
        int cur_arg;
        struct __func_args *next;
        struct __func_args *prev;
} FUNCARGS;
typedef FUNCARGS* LPFUNCARGS;

extern LPFUNCARGS pFirstFuncArg;
extern LPFUNCARGS pLastFuncArg;

extern LPFILESTATE pFirstFile;
extern LPFILESTATE pLastFile;
extern LPFILESTATE pThisFile;

//main.c
extern void SysWarn(char *);
extern void SysErr(char *);

//hash.c
extern unsigned long strhash(char *);

//parse.c
extern int parse(FILE*,FILE*);

//interp.c
extern long interpret_command(long,LPSAFELISTITEM,FILE*,FILE*);
extern long tiny_func_run(char *,FILE*,FILE*,LPSAFELISTITEM);

//string.c
extern int strip_whitespace(LPSAFESTRING);
extern char *getstrptr(LPSAFESTRING,int);
extern LPSAFESTRING setlength(LPSAFESTRING,int);
extern int zerostring(LPSAFESTRING);
extern int addrawstr(LPSAFESTRING,char *);
extern int addsafestr(LPSAFESTRING,LPSAFESTRING);
extern int addrawstrn(LPSAFESTRING,char *,int);
extern int addsafestrn(LPSAFESTRING,LPSAFESTRING,int);
extern int addsafestro(LPSAFESTRING,LPSAFESTRING,int);
extern int addsafestrno(LPSAFESTRING,LPSAFESTRING,int,int);
extern int killstring(LPSAFESTRING);
extern int replacestr(LPSAFESTRING,int,int,char *);
extern char *ltrim(char *);
extern char *ltoa(long);
extern char *ftoa(double);

//the safelistitem stuff from string.c
extern int kill_entire_list(LPSAFELISTITEM);
extern int kill_list_item(LPSAFELISTITEM);
extern LPSAFELISTITEM create_list_item(LPSAFELISTITEM);
extern int update_list_item(LPSAFELISTITEM,char *);
extern int update_list_item_from_safe_str(LPSAFELISTITEM,LPSAFESTRING);
extern LPSAFELISTITEM dup_list_item(int,int);

//and from funclist.c
extern int kill_entire_func_list(LPFUNCTIONDEF);
extern int kill_func_list_item(LPFUNCTIONDEF);
extern int kill_vars_at_depth(int);
extern LPFUNCTIONDEF create_user_function(char *,unsigned long);
extern LPFUNCTIONDEF create_intrinsic_function(char *,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*));
extern LPFUNCTIONDEF create_intrinsic_keyword(char *,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*));

extern int SetStrRet(char *);
extern int SetRealRet(float);
extern int SetLongRet(long);


//uservars.c
extern int create_user_variables(char *,LPSAFELISTITEM,int,int);
extern int eval_expression(char *,unsigned long *,long *,double *,int,int);
extern LPFUNCTIONDEF GetUserIdentifier(char *);
extern int InsertFunctionCall(char *,char *,int);

//stack.c
extern int init_stack(STACK *);
extern int push(STACK *,long);
extern long pop(STACK *);
extern int reset_stack_to_level(int);

//filestate.c

extern LPFILESTATE create_file(LPFILESTATE);
extern int kill_file(LPFILESTATE);
extern int kill_all_files(LPFILESTATE);

//funcargs.c

extern int kill_arg_list_item(LPFUNCARGS);
extern int kill_entire_arg_stack(LPFUNCARGS);
extern int create_arg_list_item();
extern int get_arg_by_num(int,int,int*,char**,int*);
extern int get_next_arg(int,int*,char**,int*);
extern int get_arg_count(int,int*);
extern char *get_arg_string(int);
extern int add_arg_string(char *);
extern int pop_arg_string();
extern int GetStringArg (char*,LPSAFESTRING);
extern int GetLongArg (char*,long*);

extern SAFESTRING str_func_arg;

//labels.c

extern int kill_all_labels(LPLABELLISTITEM);
extern int kill_labels_at_depth(int);
extern int create_label(char *,unsigned long int,unsigned long int,int);
extern int find_label(char *,unsigned long *,LPSAFELISTITEM);

//dlls.c

extern int kill_all_dlls();
extern int create_dll_entry(char *);

//compare.c

extern long EvaluateComparison(char *,FILE*,FILE*,LPSAFELISTITEM);
extern long DoCompareWork(char *,long *);
extern long DoStringCompare(char *,long *);
                                  
//and the massive list of user funcs (userfuncs.c)

extern int do_print_work(char*,LPSAFESTRING);
extern int print_eval_error(int);
extern int print_arg_error(int);
extern void init_user_func_list(void);

extern LangFunc(__user_print_func); //print
extern LangFunc(__user_eval_expression); //evaluate equation
extern LangFunc(__user_int_func); //create integer
extern LangFunc(__user_str_func); //create string.
extern LangFunc(__user_uint_func); //create unsigned
extern LangFunc(__user_float_func); //create floating point val.
extern LangFunc(__user_include_func); //include file in-place.
extern LangFunc(__user_func_creation); //allow user functions.
extern LangFunc(__user_endf_func); //end function.
extern LangFunc(__user_eval_string); //eval a str equation.

extern int do_string_eval_work(LPSAFESTRING,LPSAFESTRING);

typedef struct __Dll_Master_Pass_Info
{
	int ver;
	FUNCTIONDEF *pFirstFunc;	//the great symbol list.
	FUNCTIONDEF *pLastFunc;
	LPLABELLISTITEM pFirstLabel;	//the label list
	LPLABELLISTITEM pLastLabel;
	LPSAFELISTITEM pFirstItem;	//the function call list-out
	LPSAFELISTITEM pLastItem;	
	long int *global_cur_line;	//the line number we're on.
	long int *global_cur_user_line; //(needed for error macros)
	double *global_double_return;	//needed for math handling.
	LPFUNCARGS pFirstFuncArg;	//function arguments
	LPFUNCARGS pLastFuncArg;
	int *sub_line_goto;		//needed for redirection forcing.
	STACK *calls;			//the stack.
	int *CallDepth;			//needed for arg passing
		//and now, the function list 
	unsigned long (*strhash)(char*);
	int (*stripwhite)(LPSAFESTRING);
	char* (*getstrptr)(LPSAFESTRING,int);
	int (*zerostring)(LPSAFESTRING);
	int (*addrawstr)(LPSAFESTRING,char *);
	int (*addsafestr)(LPSAFESTRING,LPSAFESTRING);
	int (*addrawstrn)(LPSAFESTRING,char *,int);
	int (*addsafestrn)(LPSAFESTRING,LPSAFESTRING,int);
	int (*addsafestro)(LPSAFESTRING,LPSAFESTRING,int);
	int (*addsafestrno)(LPSAFESTRING,LPSAFESTRING,int,int);
	int (*killstr)(LPSAFESTRING);
	int (*replacestr)(LPSAFESTRING,int,int,char *);
	char *(*ltrim)(char *);
	char *(*ltoa)(long);
	char *(*ftoa)(double);
	LPFUNCTIONDEF (*create_intrinsic)(char *,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*));
	int (*__emu_create_user_variables)(char *,LPSAFELISTITEM,int,int);
	int (*__emu_eval_expression)(char *,unsigned long *,long *,double *,int,int);
	LPFUNCTIONDEF (*__emu_internal_GetUserIdentifier)(char *);
	long (*__emu_pop)(STACK *);
	int (*__emu_push)(STACK *,long);
	int (*__emu_kill_arg_list_item)(LPFUNCARGS);
	int (*__emu_create_arg_list_item)();
	int (*__emu_get_arg_by_num)(int,int,int*,char**,int*);
	int (*__emu_get_next_arg)(int,int*,char**,int*);
	int (*__emu_get_arg_count)(int,int*);
	char* (*__emu_get_arg_string)(int);
	int (*__emu_add_arg_string)(char *);
	int (*__emu_pop_arg_string)();
	int (*__emu_print_eval_error)(int);
	int (*__emu_print_arg_error)(int);				
        void (*__syswarn)(char*);
        void (*__syserr)(char*);
	int (*__setrealret)(float);
	int (*__setlongret)(long);
	int (*__setstrret)(char*);
} LIBRARYHANDLES;
typedef LIBRARYHANDLES* LPLIBRARYHANDLES;


#endif
