/* emusdk.c
 * (c) Copyright 2001 Aaron Kimball
 * Contains the DLL required to be linked against by
 * all library dlls; 
 *
 * RegisterLibrary() sets up the pointers back into the 
 * instance of emu required to do anything useful.
 *
 */


#include "emusdkint.h"

#define DEF_SDK_VER 1


	FUNCTIONDEF *pFirstFunc;	//the great symbol list.
	FUNCTIONDEF *pLastFunc;
	LPLABELLISTITEM pFirstLabel;	//the label list
	LPLABELLISTITEM pLastLabel;
	LPSAFELISTITEM pFirstItem;	//the function call list-out
	LPSAFELISTITEM pLastItem;	
	long int *global_cur_line;	//the line number we're on.
	long int *global_cur_user_line; //(needed for error macros)
	double *global_double_return;	//needed for math handling.
	LPFUNCARGS pFirstFuncArg;	//function arguments
	LPFUNCARGS pLastFuncArg;
	int *sub_line_goto;		//needed for redirection forcing.
	STACK *calls;			//the stack.
	int *CallDepth;	

 unsigned long (*__emu_internal_strhash)(char *);
 int (*__emu_internal_strip_whitespace)(LPSAFESTRING);
 char* (*__emu_internal_getstrptr)(LPSAFESTRING,int);
 int (*__emu_internal_zerostring)(LPSAFESTRING);
 int (*__emu_internal_addrawstr)(LPSAFESTRING,char *);
 int (*__emu_internal_addsafestr)(LPSAFESTRING,LPSAFESTRING);
 int (*__emu_internal_addrawstrn)(LPSAFESTRING,char *,int);
 int (*__emu_internal_addsafestrn)(LPSAFESTRING,LPSAFESTRING,int);
 int (*__emu_internal_addsafestro)(LPSAFESTRING,LPSAFESTRING,int);
 int (*__emu_internal_addsafestrno)(LPSAFESTRING,LPSAFESTRING,int,int);
 int (*__emu_internal_killstring)(LPSAFESTRING);
 int (*__emu_internal_replacestr)(LPSAFESTRING,int,int,char *);
 char* (*__emu_internal_ltrim)(char *);
 char* (*__emu_internal_ltoa)(long);
 char* (*__emu_internal_ftoa)(double);
 LPFUNCTIONDEF (*__emu_create_intrinsic_function)
	(char *,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*));
 int (*__emu_create_user_variables)(char *,LPSAFELISTITEM,int,int);
 int (*__emu_eval_expression)(char *,unsigned long *,long *,double *,int,int);
 LPFUNCTIONDEF (*__emu_internal_GetUserIdentifier)(char *);
 int (*__emu_push)(STACK *,long);
 long (*__emu_pop)(STACK *);
 int (*__emu_kill_arg_list_item)(LPFUNCARGS);
 int (*__emu_create_arg_list_item)();
 int (*__emu_get_arg_by_num)(int,int,int*,char**,int*);
 int (*__emu_get_next_arg)(int,int*,char**,int*);
 int (*__emu_get_arg_count)(int,int*);
 char* (*__emu_get_arg_string)(int);
 int (*__emu_add_arg_string)(char *);
 int (*__emu_pop_arg_string)();
 int (*__emu_print_eval_error)(int);
 int (*__emu_print_arg_error)(int);
        void (*__emu_syswarn)(char*);
        void (*__emu_syserr)(char*);
int (*__emu_setrealret)(float);
int (*__emu_setlongret)(long);
int (*__emu_setstrret)(char*);


int GetLibVersion()
{
	return DEF_SDK_VER;
}

int InitLibrary(LPLIBRARYHANDLES hnds)
{
	
	//this function sets us up the pointers.
	int sdk_version = DEF_SDK_VER;


	if(hnds->ver >= 1)
	{
		pFirstFunc = hnds->pFirstFunc;
		pLastFunc = hnds->pLastFunc;
		pFirstLabel = hnds->pFirstLabel;
		pLastLabel = hnds->pLastLabel;
		pFirstItem = hnds->pFirstItem;
		pLastItem = hnds->pLastItem;
		global_cur_line = hnds->global_cur_line;
		global_cur_user_line = hnds->global_cur_user_line;
		global_double_return = hnds->global_double_return;
		pFirstFuncArg = hnds->pFirstFuncArg;
		pLastFuncArg = hnds->pLastFuncArg;
		sub_line_goto = hnds->sub_line_goto;
		calls = hnds->calls;
		CallDepth = hnds->CallDepth;

		__emu_internal_strhash = hnds->strhash;
		__emu_internal_strip_whitespace = hnds->stripwhite;
		__emu_internal_getstrptr = hnds->getstrptr;
		__emu_internal_zerostring = hnds->zerostring;
		__emu_internal_addrawstr = hnds->addrawstr;
		__emu_internal_addrawstrn = hnds->addrawstrn;
		__emu_internal_addsafestr = hnds->addsafestr;
		__emu_internal_addsafestrn = hnds->addsafestrn;
		__emu_internal_addsafestro = hnds->addsafestro;
		__emu_internal_addsafestrno = hnds->addsafestrno;
		__emu_internal_killstring = hnds->killstr;
		__emu_internal_replacestr = hnds->replacestr;
		__emu_internal_ltrim = hnds->ltrim;
		__emu_internal_ltoa = hnds->ltoa;
		__emu_internal_ftoa = hnds->ftoa;
		__emu_create_intrinsic_function = hnds->create_intrinsic;
		__emu_create_user_variables = hnds->__emu_create_user_variables;
		__emu_eval_expression = hnds->__emu_eval_expression;
		__emu_internal_GetUserIdentifier = hnds->__emu_internal_GetUserIdentifier;
		__emu_pop = hnds->__emu_pop;
		__emu_push = hnds->__emu_push;
		__emu_kill_arg_list_item = hnds->__emu_kill_arg_list_item;
		__emu_create_arg_list_item = hnds->__emu_create_arg_list_item;
		__emu_get_arg_by_num = hnds->__emu_get_arg_by_num;
		__emu_get_next_arg = hnds->__emu_get_next_arg;
		__emu_get_arg_count = hnds->__emu_get_arg_count;
		__emu_get_arg_string = hnds->__emu_get_arg_string;
		__emu_add_arg_string = hnds->__emu_add_arg_string;
		__emu_pop_arg_string = hnds->__emu_pop_arg_string;
		__emu_print_eval_error = hnds->__emu_print_eval_error;
		__emu_print_arg_error = hnds->__emu_print_arg_error;
		__emu_syswarn = hnds->__syswarn;
		__emu_syserr = hnds->__syserr;        
		__emu_setrealret = hnds->__setrealret;
		__emu_setlongret = hnds->__setlongret;
		__emu_setstrret = hnds->__setstrret;

	}
	
	return 0;
}

