/* emusdk.h
 * Master include file for the 
 * Emu library SDK
 *
 * (c) Copyright 2001 Aaron Kimball
 *
 */

#ifndef __EMU_STANDARD_INCLUDE
#define __EMU_STANDARD_INCLUDE

#include<stdio.h>
#include<string.h>
#include<memory.h>
#include<ctype.h>
#include<stdlib.h>

#ifdef TARGET_LINUX
#include<dlfcn.h>
#endif

#ifdef TARGET_DOS
#include<conio.h>
#endif


typedef struct 
{
        char *str;
        long len;
} SAFESTRING;

typedef SAFESTRING* LPSAFESTRING;

#define STACK_MAX 32768
typedef struct __num_stack
{
        long data[STACK_MAX];
        int n;
} STACK;

extern STACK *calls;

typedef struct __func_args
{
        LPSAFESTRING argstr;
        int cur_arg;
        struct __func_args *next;
        struct __func_args *prev;
} FUNCARGS;
typedef FUNCARGS* LPFUNCARGS;

extern LPFUNCARGS pFirstFuncArg;
extern LPFUNCARGS pLastFuncArg;

typedef struct __safe_str_list
{
	char *str;
	long len;
	long line_no;
	long sub_line_no;
	long file_no;
        long user_line_no;
	int type; //1 for code, 0 for text, 2 for computer-generated code.
	struct __safe_str_list *next;
        struct __safe_str_list *prev;
} SAFELISTITEM;
typedef SAFELISTITEM* LPSAFELISTITEM;

extern SAFELISTITEM *pFirstItem, *pLastItem;

#define LangFunc(x) int x(char *str,FILE *fpin, FILE *fpout,LPSAFELISTITEM pThisLine,unsigned long *force_jump)

typedef struct __func_def_list
{
        unsigned long hash;
        int type; //0  for intrinsic,
                  //1  for userdefined func,
                  //2  for float,
                  //3  for int,
                  //4  for uint
                  //5  for a string.
                  //6  is internal int,
                  //7  is internal float
                  //8  is float array
                  //9  is int array
                  //10 is uint array
                  //11 is string array.
		  //12 is internal non-comparative float
        union
        {
                unsigned long line; //what line is it on?
                int (*func)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long *); //the func itself.
        } ptr;
        union
        {
                double float_val;
                long long_val;
                unsigned long ulong_val;
                LPSAFESTRING str_val;
        } val;
        int children;
        int domain; //local only? (0 = global)
	char name[41];
        struct __func_def_list *next;
        struct __func_def_list *parent; //array items point to the array.
} FUNCTIONDEF;
typedef FUNCTIONDEF* LPFUNCTIONDEF;

extern FUNCTIONDEF *pFirstFunc, *pLastFunc;

typedef struct __label_list_struct
{
	unsigned long int hash;
	unsigned long int line;
	unsigned long int subline;
	unsigned long int depth;	//function call nest depth
	unsigned long int if_depth;	//how many if-blocks it's inside.
	SAFESTRING name;
	struct __label_list_struct *pPrev;
	struct __label_list_struct *pNext;
} LABELLISTITEM;
typedef LABELLISTITEM* LPLABELLISTITEM;

extern LPLABELLISTITEM pFirstLabel;
extern LPLABELLISTITEM pLastLabel;

typedef struct __Dll_Master_Pass_Info
{
	int ver;
	FUNCTIONDEF *pFirstFunc;	//the great symbol list.
	FUNCTIONDEF *pLastFunc;
	LPLABELLISTITEM pFirstLabel;	//the label list
	LPLABELLISTITEM pLastLabel;
	LPSAFELISTITEM pFirstItem;	//the function call list-out
	LPSAFELISTITEM pLastItem;	
	long int *global_cur_line;	//the line number we're on.
	long int *global_cur_user_line; //(needed for error macros)
	double *global_double_return;	//needed for math handling.
	LPFUNCARGS pFirstFuncArg;	//function arguments
	LPFUNCARGS pLastFuncArg;
	int *sub_line_goto;		//needed for redirection forcing.
	STACK *calls;			//the stack.
	int *CallDepth;			//needed for arg passing.
		//and now, the function list 
	unsigned long (*strhash)(char*);
	int (*stripwhite)(LPSAFESTRING);
	char* (*getstrptr)(LPSAFESTRING,int);
	int (*zerostring)(LPSAFESTRING);
	int (*addrawstr)(LPSAFESTRING,char *);
	int (*addsafestr)(LPSAFESTRING,LPSAFESTRING);
	int (*addrawstrn)(LPSAFESTRING,char *,int);
	int (*addsafestrn)(LPSAFESTRING,LPSAFESTRING,int);
	int (*addsafestro)(LPSAFESTRING,LPSAFESTRING,int);
	int (*addsafestrno)(LPSAFESTRING,LPSAFESTRING,int,int);
	int (*killstr)(LPSAFESTRING);
	int (*replacestr)(LPSAFESTRING,int,int,char *);
	char *(*ltrim)(char *);
	char *(*ltoa)(long);
	char *(*ftoa)(double);
	LPFUNCTIONDEF (*create_intrinsic)(char *,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*));
	int (*__emu_create_user_variables)(char *,LPSAFELISTITEM,int,int);
	int (*__emu_eval_expression)(char *,unsigned long *,long *,double *,int,int);
	LPFUNCTIONDEF (*__emu_internal_GetUserIdentifier)(char *);
	long (*__emu_pop)(STACK *);
	int (*__emu_push)(STACK *,long);
	int (*__emu_kill_arg_list_item)(LPFUNCARGS);
	int (*__emu_create_arg_list_item)();
	int (*__emu_get_arg_by_num)(int,int,int*,char**,int*);
	int (*__emu_get_next_arg)(int,int*,char**,int*);
	int (*__emu_get_arg_count)(int,int*);
	char* (*__emu_get_arg_string)(int);
	int (*__emu_add_arg_string)(char *);
	int (*__emu_pop_arg_string)();
	int (*__emu_print_eval_error)(int);
	int (*__emu_print_arg_error)(int);				
	void (*__syswarn)(char*);
	void (*__syserr)(char*);
        int (*__setrealret)(float);
        int (*__setlongret)(long);
        int (*__setstrret)(char*);
} LIBRARYHANDLES;
typedef LIBRARYHANDLES* LPLIBRARYHANDLES;

extern int InitLibrary(LPLIBRARYHANDLES);
extern int GetLibVersion();

extern int *CallDepth;
extern int *sub_line_goto;
extern long int *global_cur_line;
extern long int *global_cur_user_line;
extern double *global_double_return;


//hash.c
extern unsigned long (*__emu_internal_strhash)(char *);
#define strhash (*__emu_internal_strhash)

//string.c
extern int (*__emu_internal_strip_whitespace)(LPSAFESTRING);
#define strip_whitespace (*emu_internal_strip_whitespace)

extern char* (*__emu_internal_getstrptr)(LPSAFESTRING,int);
#define getstrptr (*__emu_internal_getstrptr)

extern int (*__emu_internal_zerostring)(LPSAFESTRING);
#define zerostring (*__emu_internal_zerostring)

extern int (*__emu_internal_addrawstr)(LPSAFESTRING,char *);
#define addrawstr (*__emu_internal_addrawstr)

extern int (*__emu_internal_addsafestr)(LPSAFESTRING,LPSAFESTRING);
#define addsafestr (*__emu_internal_addsafestr)

extern int (*__emu_internal_addrawstrn)(LPSAFESTRING,char *,int);
#define addrawstrn (*__emu_internal_addrawstrn)

extern int (*__emu_internal_addsafestrn)(LPSAFESTRING,LPSAFESTRING,int);
#define addsafestrn (*__emu_internal_addsafestrn)

extern int (*__emu_internal_addsafestro)(LPSAFESTRING,LPSAFESTRING,int);
#define addsafestro (*__emu_internal_addsafestro)

extern int (*__emu_internal_addsafestrno)(LPSAFESTRING,LPSAFESTRING,int,int);
#define addsafestrno (*__emu_internal_addsafestrno)

extern int (*__emu_internal_killstring)(LPSAFESTRING);
#define killstring (*__emu_internal_killstring)

extern int (*__emu_internal_replacestr)(LPSAFESTRING,int,int,char *);
#define replacestr (*__emu_internal_replacestr)

extern char*(*__emu_internal_ltrim)(char *);
#define ltrim (*__emu_internal_ltrim)

extern char* (*__emu_internal_ltoa)(long);
#define ltoa (*__emu_internal_ltoa)

extern char* (*__emu_internal_ftoa)(double);
#define ftoa (*__emu_internal_ftoa)


//and from funclist.c
extern LPFUNCTIONDEF (*__emu_create_intrinsic_function)
	(char *,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*));
#define create_intrinsic_function (*__emu_create_intrinsic_function)

extern int (*__emu_setrealret)(float);
#define SetRealRet (*__emu_setrealret);

extern int (*__emu_setlongret)(long);
#define SetLongRet (*__emu_setlongret);

extern int (*__emu_setstrret)(char*);
#define SetStrRet (*__emu_setstrret);

//uservars.c
extern int (*__emu_create_user_variables)(char *,LPSAFELISTITEM,int,int);
#define create_user_variables (*__emu_create_user_variables)

extern int (*__emu_eval_expression)(char *,unsigned long *,long *,double *,int,int);
#define eval_expression (*__emu_eval_expression)

extern LPFUNCTIONDEF (*__emu_internal_GetUserIdentifier)(char *);
#define GetUserIdentifier (*__emu_internal_GetUserIdentifier)


//stack.c
extern int (*__emu_push)(STACK *,long);
#define push (*__emu_push)

extern long (*__emu_pop)(STACK *);
#define pop (*__emu_pop)

//funcargs.c

extern int (*__emu_kill_arg_list_item)(LPFUNCARGS);
#define kill_arg_list_item (*__emu_kill_arg_list_item)

extern int (*__emu_create_arg_list_item)();
#define create_arg_list_item (*__emu_create_arg_list_item)

extern int (*__emu_get_arg_by_num)(int,int,int*,char**,int*);
#define get_arg_by_num (*__emu_get_arg_by_num)

extern int (*__emu_get_next_arg)(int,int*,char**,int*);
#define get_next_arg  (*__emu_get_next_arg)

extern int (*__emu_get_arg_count)(int,int*);
#define get_arg_count (*__emu_get_arg_count)

extern char* (*__emu_get_arg_string)(int);
#define get_arg_string (*__emu_get_arg_string)

extern int (*__emu_add_arg_string)(char *);
#define add_arg_string (*__emu_add_arg_string)

extern int (*__emu_pop_arg_string)();
#define pop_arg_string (*__emu_pop_arg_string)

//userfuncs.c

extern int (*__emu_print_eval_error)(int);
#define print_eval_error (*__emu_print_eval_error)

extern int (*__emu_print_arg_error)(int);
#define print_arg_error (*__emu_print_arg_error)

//main.c
extern void (*__emu_syswarn)(char*);
#define SysWarn (*__emu_syswarn)

extern void (*__emu_syserr)(char*);
#define SysErr (*__emu_syserr)


#endif
