/* funclist.c
 * HTMI html macro interpretter
 * 
 * contains a linked-list which keeps track of functions 
 * for use by the htmi program, by hash value of the funcname.
 *
 * (c) Copyright 2000 Aaron Kimball
 *
 */

#include<stdio.h>
#include<string.h>
#include<memory.h>


LPFUNCTIONDEF pFirstFunc;
LPFUNCTIONDEF pLastFunc;

int kill_vars_at_depth(int depth)
{
	LPFUNCTIONDEF lpNext;
	LPFUNCTIONDEF lpThis;
	LPFUNCTIONDEF lpPrev;

	lpPrev = 0;
	lpThis = pFirstFunc;
	while(lpThis)
	{

		lpNext = lpThis->next;
		if(lpThis->domain == depth) 
		{
			if(lpPrev) lpPrev->next = lpNext;
			if(lpThis == pLastFunc) pLastFunc = lpPrev;
			kill_func_list_item(lpThis);
		}
		else
		{
			lpPrev = lpThis;
		}
		lpThis = lpNext;
	}


	return 0;
}

int kill_entire_func_list(LPFUNCTIONDEF lpFirst)
{
	LPFUNCTIONDEF lpNext;
	LPFUNCTIONDEF lpThis;

	lpThis = lpFirst;
	while(lpThis)
	{
		lpNext = lpThis->next;
		kill_func_list_item(lpThis);
		lpThis = lpNext;
	}

	return 0;
}

int kill_func_list_item(LPFUNCTIONDEF lpss)
{
        if(lpss->type == 5)
        {
                killstring(lpss->val.str_val);
                free(lpss->val.str_val);
		lpss->val.str_val = 0;
        }
	free(lpss);
	return 0;
}

LPFUNCTIONDEF create_intrinsic_function(char *funcname,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*))
{
	LPFUNCTIONDEF lpCur;
        LPFUNCTIONDEF lpPrev = pLastFunc;
	unsigned long hash;

	hash = strhash(funcname);

	lpCur = calloc(sizeof(FUNCTIONDEF),1);
        lpCur->next = 0;
        lpCur->children = 0;
	lpCur->hash = hash;
	lpCur->type = 0; //intrinsic
	lpCur->ptr.func = ptr;
	memset(lpCur->name,0,41);
	strncpy(lpCur->name,funcname,40);

        if(lpPrev) { lpPrev->next = lpCur; }
	else { pFirstFunc = lpCur; }
	pLastFunc = lpCur;

	return lpCur;
}

LPFUNCTIONDEF create_intrinsic_keyword(char *funcname,int (*ptr)(char*,FILE*,FILE*,LPSAFELISTITEM,unsigned long*))
{
	LPFUNCTIONDEF lpCur;
        LPFUNCTIONDEF lpPrev = pLastFunc;
	unsigned long hash;

	hash = strhash(funcname);

	lpCur = calloc(sizeof(FUNCTIONDEF),1);
        lpCur->next = 0;
        lpCur->children = 0;
	lpCur->hash = hash;
	lpCur->type = 13; //intrinsic keyword
	lpCur->ptr.func = ptr;
	memset(lpCur->name,0,41);
	strncpy(lpCur->name,funcname,40);

        if(lpPrev) { lpPrev->next = lpCur; }
	else { pFirstFunc = lpCur; }
	pLastFunc = lpCur;

	return lpCur;
}
LPFUNCTIONDEF create_user_function(char *funcname,unsigned long ptr)
{
	LPFUNCTIONDEF lpCur;
        LPFUNCTIONDEF lpPrev = pLastFunc;
	unsigned long hash;

	hash = strhash(funcname);

	lpCur = calloc(sizeof(FUNCTIONDEF),1);
        lpCur->next = 0;
        lpCur->children = 0;
	lpCur->hash = hash;
	lpCur->type = 1; //user defined
	lpCur->ptr.line = ptr;
	memset(lpCur->name,0,41);
	strncpy(lpCur->name,funcname,40);

        if(lpPrev) lpPrev->next = lpCur;
	pLastFunc = lpCur;

	return lpCur;
}

int SetStrRet(char *str)
{
	//sets the ret variable to 'str'
	LPFUNCTIONDEF lpFunc;

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type != 5)
	{
		lpFunc->type = 5;
		lpFunc->val.str_val = malloc(sizeof(LPSAFESTRING));
	}

	zerostring(lpFunc->val.str_val);
	addrawstr(lpFunc->val.str_val,str);

	return 0;
}

int SetRealRet(float value)
{
	//sets the ret variable to 'value'

	LPFUNCTIONDEF lpFunc;

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type == 5)
	{
		killstring(lpFunc->val.str_val);
		free(lpFunc->val.str_val);
	}
	lpFunc->type = 2;

	lpFunc->val.float_val = value;

	return 0;
}

int SetLongRet(long value)
{
	//sets the ret variable to 'value'

	LPFUNCTIONDEF lpFunc;

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type == 5)
	{
		killstring(lpFunc->val.str_val);
		free(lpFunc->val.str_val);
	}
	lpFunc->type = 3;

	lpFunc->val.long_val = value;

	return 0;
}

