//labels.c
//
//contains the linked-list code
//which controls the line-labels in emu.
//
// (c) 2001 Aaron Kimball

LPLABELLISTITEM pFirstLabel=0;
LPLABELLISTITEM pLastLabel=0;
SAFESTRING labelseekname;
unsigned long int labelseekhash,labeldepth;

int create_label(char *name,unsigned long int line_no,unsigned long int sub_line_no,int linetype)
{
	LPLABELLISTITEM pNew;
	LPLABELLISTITEM pNext;
	char *p1=0;
	unsigned long int thehash;
	int evil = 0;

	//first, simple validity check.
	p1=name;
	if(!isalpha(*p1) && *p1 != '_') evil = 1;
	for(p1++;*p1;p1++) if(!isalnum(*p1) && *p1 != '_') evil = 1;
	if(linetype == 2) evil = 0;
	if(evil) return 2;

	thehash = strhash(name);

	//next: check to see if its the same as any other label. (treewalk)
	pNext = pFirstLabel;
	while(pNext)
	{
		if(pNext->hash == thehash)
		{
			if(pNext->depth == CallDepth && !strcmp(pNext->name.str,name))
			{
				return 3;
			}
		}
		pNext = pNext->pNext;
	}

	//ok, its cool, plug 'er in.

	pNew = malloc(sizeof(LABELLISTITEM));
	if(!pNew) return 1; //out of memory. bomb.

	if(pLastLabel) 
	{
		pLastLabel->pNext = pNew; 
		pNew->pPrev = pLastLabel;
		pLastLabel = pNew;
	}
	else 
	{ 
		pLastLabel = pNew; 
		pLastLabel->pPrev = 0;
	}

	pNew->line = line_no;
	pNew->subline = sub_line_no;
	pNew->hash = thehash;
	pNew->depth = CallDepth;
	pNew->if_depth = IfDepth;
	pNew->pNext = 0;
	zerostring(&(pNew->name));
	addrawstr(&(pNew->name),name);

	return 0;
}

int kill_labels_at_depth(int depth)
{
	LPLABELLISTITEM pNext;
        LPLABELLISTITEM pThis;

        pNext = pFirstLabel; //start at top of the tree.

        while(pNext)
        {
                pThis=pNext;
                pNext=pThis->pNext;
		if(pThis->depth == depth)  //if we should delete it, then:
		{
			//patch the current label out of the list
			if(pThis->pNext) pThis->pNext->pPrev = pThis->pPrev; 
			if(pThis->pPrev) pThis->pPrev->pNext = pThis->pNext;
			//and destroy it
			killstring(&(pThis->name));
			free(pThis);
		}
        }

        return 0;

}

int kill_all_labels(LPLABELLISTITEM pFirst)
{
	LPLABELLISTITEM pNext;
	LPLABELLISTITEM pThis;

	pNext = pFirst;

	while(pNext)
	{
		pThis=pNext;
		pNext=pThis->pNext;
		killstring(&(pThis->name));
		free(pThis);
	}
	
	return 0;
}

int find_label(char *name,unsigned long *force_jump,LPSAFELISTITEM pThisLine)
{
        LPLABELLISTITEM pNextLabel;
        unsigned long int labelhash;

	labelhash = strhash(name);

        pNextLabel = pLastLabel; //search the tree backwards, in order to
                                 //allow for multiple recursive calls to this function.
        while(pNextLabel)
        {
                if(pNextLabel->hash == labelhash)
                {
                        if(strcmp(pNextLabel->name.str,name))
                        {
                                //not the right label.
                                pNextLabel=pNextLabel->pPrev;
                                continue;
                        }
                        if(pNextLabel->depth == CallDepth)
                        {
                                //we've found our label!
                                //to jump to it, we have to do three things:
                                //a) unload the entire stack down to the function call
                                //b) read the label's ifdepth and insert dummy stack fillers.
                                //c) set the next line to exec!

                                reset_stack_to_level(pNextLabel->if_depth);
                                *force_jump = pNextLabel->line;
                                sub_line_goto = pNextLabel->subline;
                                IfDepth = pNextLabel->if_depth;
                                CurIfEvaluation = 0;

                                return 0;
                        }
                        else
                        {
                                //error: label is at an inappropriate depth.
                                SysErr("Illegal jump out of current function context");
                                return 1;
                        }
                }

                pNextLabel=pNextLabel->pPrev;
        }


	return 1; //didn't find it.
}
