/* main.c
 * main source file for emu system
 * contains, well, main() for one...
 *
 * (c) Copyright 2000 Aaron Kimball
 *
 */

#include<stdio.h>
#include<string.h>
#include<memory.h>

//i always forget to update this. the official version
//is in the 'version' file.
char cur_version[] = "Emu v1.0";


long int global_cur_line;
long int global_cur_subline;
long int global_cur_user_line;
int safemode = 0; //normal (1 = activated; no file usage allowed)
int cur_file_num = 0;
int warning_mode = 0; //normal (1 = as error, 2 = suppressed)
int error_mode = 0; //normal (1 = bomb out)
int backgroundmode = 0; //normal (1 = is in background; user input disabled.)
int quietenv = 0; //normal (1 = silent.)

void SysErr(char *x)
{
	fprintf(fpout_global,"Error: %ld: %s\n",global_cur_user_line,x);
	if(error_mode == 1) exit(1);
}

void SysWarn(char *x) 
{
	if(warning_mode == 1) { SysErr(x); return; }
	if(warning_mode == 2) return;
	fprintf(fpout_global,"Warning: %ld: %s\n",global_cur_user_line,x);
}

void print_instructions()
{
	printf("%s\n\n",cur_version);
        printf("Usage: emu [-v | -u | -b -s -c dir -a -w -q] filename [-o outputfile]\n");
	exit(0);
}

void exit_cleanup(void)
{
	//everything we gotta do at the end of the program.

	//any EOF errors?
	if(bSeeking == 4) fprintf(stderr,"Error: %i: Label not found\n",userseekline);
	if(bSeeking == 5) fprintf(stderr,"Error: %i: Unterminated loop\n",userseekline);

	//done... let's go home.
        
	kill_entire_list(pFirstItem); //kill the user's program
        kill_entire_func_list(pFirstFunc); //kill the func hash table
        kill_entire_arg_stack(pFirstFuncArg); //function argument list.
        kill_all_files(pFirstFile); //kill the file list
	kill_all_labels(pFirstLabel); //kill any line labels.
	kill_all_dlls(); //kill any loaded libraries.

	killstring(&LoopIncAssign);
        killstring(&seekstr);
	killstring(&labelseekname);
        killstring(&str_func_arg);

}

int main(int argc,char **argv)
{
	char input_file[256];
	int i;
	FILE *fpout,*fpin;

	fpout = stdout;
	fpout_global = stdout;
	memset(input_file,0,256);

	if(argc == 1) print_instructions();

	for(i=1;i<argc;i++)
	{
		if(argv[i][0] == '-')
		{
			switch(argv[i][1])
			{
			case 'o':
outputfile:
				//output to file:
				i++;
				fpout = 0;
				if(argv[i]) fpout = fopen(argv[i],"w");
				if(!fpout)
				{
					printf("Warning: Could not redirect output.\n");
					fpout = stdout;
				}
				break;
			case 'b':
bgmode:
				//background mode.
				if(fpout == stdout) fpout = fopen("/dev/null","w");	
				if(fpout == 0) 
					SysErr("Could not open /dev/null.");
				backgroundmode = 1;
				quietenv = 1;
				break;
			case 'c':
setcwd:
				//set cwd
				i++;
				if(argv[i]) chdir(argv[i]);
				break;
			case 'a':
allscript:
				//all of the file is script; no need for <$
				bInScript = 1;
				break;
			case 'w':
nowarns:
				//warnings suppressed
				warning_mode = 2;
				break;
			case 's':
				//strict mode.
strict:
				error_mode = 1;
				warning_mode = 1;
				break;
			case 'q':
quiet_environ:
				quietenv = 1;
				break;
			case 'v':
ver_info:
				printf("%s\n\n",cur_version);
				exit(0);
				break;
			case 'u':
usage_info:
				print_instructions();
				break;
			case 0:
				//the entire parameter is just "-"
				//thus, just read from stdin.
				strcpy(input_file,"::/stdin");
				break;
			case '-':
				//we've got a long switch name.
				if(strcmp(argv[i]+2,"output")==0) goto outputfile;
				if(strcmp(argv[i]+2,"background")==0) goto bgmode;
				if(strcmp(argv[i]+2,"strict")==0) goto strict;
				if(strcmp(argv[i]+2,"all")==0) goto allscript;
				if(strcmp(argv[i]+2,"curdir")==0) goto setcwd;
				if(strcmp(argv[i]+2,"no-warnings")==0) goto nowarns;
				if(strcmp(argv[i]+2,"quiet-input")==0) goto quiet_environ;
				if(strcmp(argv[i]+2,"safe")==0) safemode = 1;
				if(strcmp(argv[i]+2,"version")==0) goto ver_info;
				if(strcmp(argv[i]+2,"usage")==0) goto usage_info;
			default:
				printf("Unrecognized switch \'%s\'\n",argv[i]);
				break;
			} //possible switches
		} //if switch
		else
		{
			//you're the input file!
			strcpy(input_file,argv[i]);
		}
	} //loop thru args

	
	if(!strlen(input_file)) print_instructions();

	if(strcmp(input_file,"::/stdin")==0) 
	{
		fpin = stdin;
		if(!quietenv)
		{
		        printf("%s\n",cur_version);
			printf("Emu command interpretter by Aaron Kimball (c) 2001\n\n");
		}
	}
	else
	{
		fpin = fopen(input_file,"r");
		if(!fpin) 
		{
			printf("Error: Could not open input file.\n");
			return 0;
		}
	}

	bCopying = 1;

        pFirstFile = create_file(0);
        pFirstFile->fp = fpin;
        pFirstFile->bCopying = 1;
        pFirstFile->file_num = 0;
        pFirstFile->line_num = 1;

	pFirstItem = create_list_item(0); //create the first element of the command array;
	pLastItem = pFirstItem;

        init_user_func_list(); //put together the list of funcs the user can call

        init_stack(&calls);

	zerostring(&LoopIncAssign);
        zerostring(&seekstr);
	zerostring(&labelseekname);
        zerostring(&str_func_arg);

	//schedule 'cleanup' commands
	atexit(exit_cleanup);

        create_arg_list_item();
        pFirstFuncArg = pLastFuncArg;

        //TODO: Now is the time for us to pass args to the 'main'.
        //if we wish to do so. (on pFirstFuncArg)

	//Ok, go for it! :)
	parse(fpin,fpout); //it's time to rock and roll (touche, mr. scullin)

	//after parsing is complete, the show's over. 
	return 0;
}
