/* parse.c
 * HTMI html macro interpretter
 * 
 * this contains...well...the parsing engine.
 * parse() reads in the file, echoes back out regular text,
 * and excizes the script parts..
 * takes them, and then get_one_statement() pulls out individual
 * commands.
 *
 * (c) Copyright 2000 Aaron Kimball
 *
 */

#include<stdio.h>
#include<string.h>
#include<memory.h>
#include<ctype.h>

int bInFunc=0;
int bAbort=0;
int bCopying=1;
int bInScript=0;
long next_statement=0; //next statement number in the queue for adds
long next_user_line_number=0;
long next_to_exec=-1;  //next statement we actually need to exec. '-1' for 'read another'
SAFELISTITEM *pFirstItem, *pLastItem;

//pull out a statement from the current line
//and grab part of next line(s) if needed to complete the statement
//returns the unused portion
int get_one_statement(LPSAFESTRING curline,FILE *fpin,FILE *fpout,int def_ret)
{
        char *p1=0,*p2=0,*p3=0,*p4=0,*p5=0; //yikes!
	int getwholeline = 0;
        SAFESTRING str_temp;
        SAFESTRING str_curline;
        SAFESTRING str_statement;
        LPSAFELISTITEM pNextItem;
	char nextline[1024];
	int offset;
        int still_in_script = def_ret;
        int num_quotes=0;
        int numslashes=0;
        char *n1=0;

//printf("** INPUT: [[%s]]\n",curline->str);
        zerostring(&str_statement);
	zerostring(&str_curline);

        addrawstr(&str_curline,curline->str);

	p1 = getstrptr(&str_curline,0);
	offset = 0;

	//advance past whitespace...
        while(*p1 != 0) if(!isalnum(*p1) && *p1 != '_' && *p1 != '#') { p1++; offset++; } else { break; }
        if(!strlen(p1))
        {
                killstring(curline);
                zerostring(curline);
                return still_in_script; //bomb if no string left ;) (read: blank line or whatnot)
        }

	if(*p1 == '#') 
	{
		//its a comment. don't parse it, just find it all (it goes to the \n) 
		//and remove it from the in-feed string.

                p2 = strchr(p1,'\n');
		if(p2) offset = p2 - getstrptr(&str_curline,0)+1; 

	        killstring(curline);
	        zerostring(curline);
	        if(strlen(p2+1)) addsafestro(curline,&str_curline,offset); 

		killstring(&str_curline);
		killstring(&str_statement);

	        return still_in_script; //bye
	}

	p3 = strstr(str_curline.str,"$>");
	if (p3 < p1 && p3) 
	{
		//we're out of statements,
		//and the line contains text to be echoed to the screen.
		p3 += 2;
		p2 = strstr(p3,"<$"); 
		if(p2 == 0) p2 = getstrptr(&str_curline,strlen(str_curline.str));
		zerostring(&str_temp);
		addrawstrn(&str_temp,p3,p2-p3);

		//add the statement to the list..
                if(next_statement) pNextItem = create_list_item(pLastItem);
		else { pNextItem = pLastItem; }

		pLastItem = pNextItem;
		pNextItem->user_line_no = next_user_line_number;
		pNextItem->line_no = next_statement;
		pNextItem->type = 0; //text.
		update_list_item_from_safe_str(pNextItem,&str_temp);
		
		interpret_command(next_statement,pNextItem,fpin,fpout);

		killstring(&str_temp);

		//now test if there's a new script starting on this line
		//and start evaluating that.
		p1 = strstr(p3,"<$"); 
		if(p1)
		{
			//yes there is one.
			p1 += 2;
			offset = p1 - getstrptr(&str_curline,0);
		        while(*p1 != 0) if(!isalpha(*p1) && !isdigit(*p1) && *p1 != '_') { p1++; offset++; } else { break; }
		}
		else
		{
			//nope.
			//zero out the input, we're sendin' nothing back.
			killstring(curline);
			zerostring(curline);
			return 0; //and we ended the script, so stay out!
		}
	}
	p3 = 0;

	getwholeline = 0;
	//check for all keywords which reserve a full line
	//TODO: Please optimize these, i'm sure its possible.
        if(strncmp(p1,"else ",5)==0) getwholeline=1;
        if(strncmp(p1,"else\t",5)==0) getwholeline=1;
        if(strncmp(p1,"else\n",5)==0) getwholeline=1;
        if(strncmp(p1,"else\0",5)==0) getwholeline=1;
        if(strncmp(p1,"loop ",5)==0) getwholeline=1;
        if(strncmp(p1,"loop\t",5)==0) getwholeline=1;
        if(strncmp(p1,"loop\n",5)==0) getwholeline=1;
        if(strncmp(p1,"endif ",6)==0) getwholeline=1;
        if(strncmp(p1,"endif\t",6)==0) getwholeline=1;
        if(strncmp(p1,"endif\n",6)==0) getwholeline=1;
	if(strncmp(p1,"endif\0",6)==0) getwholeline=1;
        if(strncmp(p1,"if ",3)==0) getwholeline=1;
        if(strncmp(p1,"if\t",3)==0) getwholeline=1;
        if(strncmp(p1,"if\n",3)==0) getwholeline=1;
        if(strncmp(p1,"elseif ",7)==0) getwholeline=1;
        if(strncmp(p1,"elseif\t",7)==0) getwholeline=1;
        if(strncmp(p1,"for ",4)==0) getwholeline=1;
        if(strncmp(p1,"for\t",4)==0) getwholeline=1;
        if(strncmp(p1,"for\n",4)==0) getwholeline=1;
	if(strncmp(p1,"while ",6)==0) getwholeline=1;
	if(strncmp(p1,"while\t",6)==0) getwholeline=1;
	if(strncmp(p1,"while\n",6)==0) getwholeline=1;
        if(strncmp(p1,"func ",5)==0) getwholeline=1;
        if(strncmp(p1,"func\t",5)==0) getwholeline=1;
        if(strncmp(p1,"return\0",7)==0) getwholeline=1;
        if(strncmp(p1,"return\n",7)==0) getwholeline=1;
        if(strncmp(p1,"return ",7)==0) getwholeline=1;
        if(strncmp(p1,"return\t",7)==0) getwholeline=1;


	//at this point we know to either search for a '\n' or a ';'
	//so do it ;)

        p2 = p3 = 0;
	while(p2 == 0)
	{
		switch(getwholeline)
		{
		case 0:
			//get to a semicolon
			p2 = strchr(p1,';');
			break;
		case 1: 
			//get to a \n
			p2 = strchr(p1,'\n');
			break;
                } //test to see if we need to get more...

                if(!p2) p2 = strchr(p1,'$');

                //test to see if p2 is within a set of quotes.
                num_quotes = 0;
                for(p4 = p1;strlen(p4);p4++)
                {
                        //run thru the string and count quotes that are
                        //to the left of p2 (or just exist, if no p2)
                        if(*p4 == '\"' && p4 < p2)
                        {
                                numslashes=0;
                                for(n1 = p4-1;n1 > p2;n1--)
                                {
                                        if(*n1 == '\\') numslashes++;
                                        else break;
                                }
                                if(!(numslashes%2)) num_quotes++;
                        }
                }

                if(num_quotes % 2) p2 = 0; 

                if(p3) break;

		if(p2 == 0) 
		{
			//yeah, we need to grab the next line.
			memset(nextline,0,1024);
			if(feof(fpin))
			{
				printf("Error: Unterminated statement; Unexpected EOF\n");
                                return still_in_script; //poof
			}
			if(fpin == stdin && !quietenv) fprintf(fpout,">");
			fgets(nextline,1024,fpin);
			if(strchr(nextline,'\n')) next_user_line_number++;
       	                addrawstr(&str_curline,nextline);
                        p1 = getstrptr(&str_curline,0); //make sure our pointer survived any change..
                        p3 = strstr(nextline,"$>");
                        if(p3)
                        {
                            //we've hit the edge of a script.
                            //force termination at the $
                            still_in_script = 0;
                        }
		}
	} //while there isn't enuf string...

	//ok, we have it all in there..
        p5 = p1;
retest_for_end : 
	switch(getwholeline) //find the endpoint of the statement
	{
	case 0:
		//get to a semicolon
                p2 = strchr(p5,';');
                if(!p2 && p3) p2 = strchr(p5,'$');
                else p3 = 0;
		break;
	case 1: 
		//get to a \n
                p2 = strchr(p5,'\n');
                if(!p2 && p3) p2 = strchr(p5,'$');
                else p3 = 0;
		break;
	} //find the endpoint of the statement

        //make sure we have the /real/ end of line, not some quoted text..
        num_quotes = 0;
        for(p4 = p1;*p4;p4++)
        {
                //run thru the string and count quotes that are
                //to the left of p2 (or just exist, if no p2)
                if(*p4 == '\"' && p4 < p2)
                {
                        numslashes=0;
                        for(n1 = p4-1;n1 > p2;n1--)
                        {
                                if(*n1 == '\\') numslashes++;
                                else break;
                        }
                        if(!(numslashes%2)) num_quotes++;
                }
        }

        if(num_quotes % 2)
        {
                p5 = p2 + 1;
                goto retest_for_end; //eeeew a goto! (bah, it works)         
        }


        //hurrah, we've found it!
        //now that we have a lock on where our string begins and ends,
        //pull it out into str_curline
	addrawstrn(&str_statement,p1,p2-p1);
        offset += strlen(str_statement.str) + 1;
        if(p3) offset += 2;
        p1 = getstrptr(&str_curline,offset);


        //and now make the str_curline be legible to the interper
//        strip_whitespace(&str_curline);
        strip_whitespace(&str_statement);

	//add the statement to the list..
        if(next_statement) pNextItem = create_list_item(pLastItem);
	else { pNextItem = pLastItem; }

	pLastItem = pNextItem;
	pNextItem->line_no = next_statement;
	pNextItem->user_line_no = next_user_line_number;
	pNextItem->type = 1; //command, not text.
	update_list_item_from_safe_str(pNextItem,&str_statement);


	//interpret it!
	interpret_command(next_statement,pNextItem,fpin,fpout);

	//at this point, update the parent string;
	//it needs to grow to encompass any additional data we had to read in
	//and also shift along to forget about the statement we just parsed.
        killstring(curline);
        zerostring(curline);
        if(strlen(p1)) addsafestro(curline,&str_curline,offset); 

	killstring(&str_curline);
	killstring(&str_statement);

        return still_in_script; //bye
}


// the main l00p.
int parse(FILE *fpin,FILE *fpout)
{
        char curline[16384];
        char tempstr[16384];
        SAFESTRING str_line;
	SAFESTRING str_temp;
	LPSAFELISTITEM pNextItem;
	LPSAFELISTITEM pSearchItem=0;
	LPSAFELISTITEM pLastSearch=0;
        char *p1=0,*p2=0;


        zerostring(&str_line);

	while(1)
	{
		while(next_to_exec != -1) 
                {
                	pSearchItem = pLastSearch;
			if(!pSearchItem) pSearchItem = pFirstItem;
        		while(pSearchItem)
                        {
				if(pSearchItem->line_no > next_to_exec) pSearchItem = pFirstItem;
				if(pSearchItem->line_no == next_to_exec && 
				   pSearchItem->sub_line_no == sub_line_goto) break;
                                pSearchItem = pSearchItem->next;
				if(!pSearchItem) break;
			}
			pLastSearch = pSearchItem;
			interpret_command(next_to_exec,pSearchItem,fpin,fpout);
			if(bAbort)
			{
                        	killstring(&str_line);
				exit(0);         
			}
                }
		memset(curline,0,1024); //clear it
		if(feof(fpin)) break;

                //test if we need a new line..
                if(!strlen(str_line.str))
                {
initial_grab_from_file:
                        memset(curline,0,16384);
			if(fpin == stdin && !quietenv) fprintf(fpout,">");
                        fgets(curline,1024,fpin); //read a line in
			if(strchr(curline,'\n')) next_user_line_number++;
//printf("** FILEGRAB [[%s]]\n",curline);
			if(strstr(curline,"#!")==curline && strchr(curline,'\n') && next_user_line_number == 1)  
			{
				p1 = strchr(curline,'\n');
				if(!p1) goto initial_grab_from_file;
				if(strlen(p1+1) == 0) goto initial_grab_from_file; //no string left. tank out.
				memset(tempstr,0,16384);
				strcpy(tempstr,p1+1);
				memset(curline,0,16384);
				strcpy(curline,tempstr);
			}
                }
                else
                {
                        //we don't read more data cause we've read plenty
                        //dump what we read last time into current buffer..
                        memset(curline,0,16384);
                        strncpy(curline,str_line.str,16384);
//printf("*** DUMPED BACK: [[[%s]]]\n",curline);
                        bInScript = get_one_statement(&str_line,fpin,fpout,bInScript);


			while(next_to_exec != -1) 
        	        {
                		pSearchItem = pLastSearch;
				if(!pSearchItem) pSearchItem = pFirstItem;
        			while(pSearchItem)
	                        {
					if(pSearchItem->line_no > next_to_exec) pSearchItem = pFirstItem;
					if(pSearchItem->line_no == next_to_exec && 
					   pSearchItem->sub_line_no == sub_line_goto) break;
                                	pSearchItem = pSearchItem->next;
					if(!pSearchItem) break;
				}
				pLastSearch = pSearchItem;
				interpret_command(next_to_exec,pSearchItem,fpin,fpout);
				if(bAbort)
				{
                	        	killstring(&str_line);
					exit(0);         
				}
        	        }


                        continue;
                }
//printf("**INITIAL STATE: %i\n",bInScript);
		if(!bInScript)
		{
			//test to see if a script starts. if it does, 
			//copy the html before it to the stdout, then eval.
			//otherwise, copy whole line to output
                        if(!strlen(curline)) break;
			if((p1=strstr(curline,"<$")) > 0)
			{
				p2 = p1 + 2; //p2 contains script start point.
				if(p1 != curline)
				{
					//we have pre-script html to print out.
					memset(tempstr,0,1024);
					strncpy(tempstr,curline,p1-curline);

					//add the statement to the list..
                                        if(next_statement) pNextItem = create_list_item(pLastItem);
					else { pNextItem = pLastItem; }

					pLastItem = pNextItem;
					pNextItem->line_no = next_statement;
					pNextItem->user_line_no = next_user_line_number;
					pNextItem->type = 0; //text.
					update_list_item(pNextItem,tempstr);
		
                                        interpret_command(next_statement,pNextItem,fpin,fpout);
				} //if pre-script to print
				bInScript = 1;
			} //if we found a script start.
			else
			{
				//copy whole thing to screen
				//add the statement to the list..
                                if(next_statement) pNextItem = create_list_item(pLastItem);
				else { pNextItem = pLastItem; }

				pLastItem = pNextItem;
				pNextItem->line_no = next_statement;
				pNextItem->user_line_no = next_user_line_number;
				pNextItem->type = 0; //text.
				update_list_item(pNextItem,curline);
		
                                interpret_command(next_statement,pNextItem,fpin,fpout);
				continue;
			} //no script on this line.
		} //did not start out in a script..
		else
		{
			//already in a script.
			p2 = curline;
                        if(!strlen(curline)) break;
                        if(strstr(curline,"<$")==ltrim(curline)) p2 = curline + 2;
		}

		if(bInScript)
		{
			//ok! it's time to play "actually do some work!"
			//pull out a single expression (something terminated
			//by either a ';' or a newline in the case of 'if' and 'elseif', etc

			//search for an end-of-script condition
			//if one is found, 
			//the stuff before it is script. exec it.
			//then print post-script html to output.

                        if(!strlen(p2)) break;
                        if((p1=strstr(p2,"$>")) > 0)
			{
				//yeap
                                bInScript = 0;

				//grab script fragment, if exists.
				if(p1 != p2)
				{
					memset(tempstr,0,1024);
					strncpy(tempstr,p2,p1-p2);
                                        //and lets be nice.. if they forget
                                        //to terminate with \n or ;, we're
                                        //good enough to give them it.
                                        //(really, its easier for the parser..)
					strcat(tempstr,"\n;\n");
                                        killstring(&str_line);
                                        zerostring(&str_line);
                                        addrawstr(&str_line,tempstr);
//printf("** CALLING (%i)\n",bInScript);
                                        bInScript = get_one_statement(&str_line,fpin,fpout,0); 
					while(next_to_exec != -1) 
        			        {
	        	        		pSearchItem = pLastSearch;
						if(!pSearchItem) pSearchItem = pFirstItem;
        					while(pSearchItem)
	                		        {
							if(pSearchItem->line_no > next_to_exec) pSearchItem = pFirstItem;
							if(pSearchItem->line_no == next_to_exec && 
							   pSearchItem->sub_line_no == sub_line_goto) break;
        	                        		pSearchItem = pSearchItem->next;
							if(!pSearchItem) break;
						}
						pLastSearch = pSearchItem;
						interpret_command(next_to_exec,pSearchItem,fpin,fpout);
						if(bAbort)
						{
                			        	killstring(&str_line);
							exit(0);         
						}
		        	        }
				} //if script is on this line.
		
                                p1 += 2;
                                if(strlen(p1))
				{
					//okay. the line contained a $>; everything before it has 
					//been dealt with, so we need to print out any text following it.
					//the problem is merely that if another <$ opens up, we need to 
					//respect that as well.
					//p1 points to immediately after the $>.

					p2 = strstr(p1,"<$"); //p2 now points to the beginning
								//of any script, etc which follows txt on same line.
					if(!p2) p2 = p1 + strlen(p1);

					memset(tempstr,0,1024);
                                        strncpy(tempstr,p1,p2-p1);
					//add the statement to the list..
                                        if(next_statement) pNextItem = create_list_item(pLastItem);
					else { pNextItem = pLastItem; }
					pLastItem = pNextItem;
					pNextItem->line_no = next_statement;
					pNextItem->user_line_no = next_user_line_number;
					pNextItem->type = 0; //text.
					update_list_item(pNextItem,tempstr);
		
                                        interpret_command(next_statement,pNextItem,fpin,fpout);

					//okay. at this point, p2 *may* point to a <$ in curline.
					//we need to set str_line.str to that point. 
					if(p2 == strstr(p2,"<$"))
					{
						p1 = 0;
						if(strlen(str_line.str) > 16383)
						{
							zerostring(&str_temp);
							p1 = strstr(str_line.str,p2);
							p1 += strlen(p2);
							addrawstr(&str_temp,p1);
						}
						zerostring(&str_line);
						addrawstr(&str_line,p2);
						if(p1) 
						{
							addsafestr(&str_line,&str_temp);
							killstring(&str_temp);
						}
						continue;
					}
				} //if there is post-script html
			} //if script ends
			else
			{
                                //no terminator on this line, exec it normally.
                                killstring(&str_line);
                                zerostring(&str_line);
                                addrawstr(&str_line,p2);
//printf("** CALLING (%i)\n",bInScript);
                                bInScript = get_one_statement(&str_line,fpin,fpout,1);

                                while(next_to_exec != -1) 
                                {
                                        pSearchItem = pLastSearch;
					if(!pSearchItem) pSearchItem = pFirstItem;
        		                while(pSearchItem)
                                        {
						if(pSearchItem->line_no > next_to_exec) pSearchItem = pFirstItem;
						if(pSearchItem->line_no == next_to_exec && 
							pSearchItem->sub_line_no == sub_line_goto) break;
                                                pSearchItem = pSearchItem->next;
						if(!pSearchItem) break;
                                        }
                                        pLastSearch = pSearchItem;
                                        interpret_command(next_to_exec,pSearchItem,fpin,fpout);
                                        if(bAbort)
                                        {
                                                killstring(&str_line);
                                                exit(0);
                                        }
                                }

			}
		} //if in a script.
	} //main while loop.

        if(bSeeking == 1)
        {
		//we've hit the end of the file while searching for a function
                fprintf(stderr,"Error: %i: Identifier (%s) not defined.\n",userseekline,seekstr.str);
        }

        killstring(&str_line);

	return 0; //i don't want to be your fool...bye bye bye!
} 
