/* string.c
 * HTMI html macro interpretter
 * 
 * contains a safe_string "class"
 * that allows the string to grow as needed.
 *
 * (c) Copyright 2000 Aaron Kimball
 *
 */

#include<stdio.h>
#include<string.h>
#include<memory.h>

char *ftoa(double d)
{
        static char bob[64];
        memset(bob,0,64);

        sprintf(bob,"%f",d);

        return bob;
}

char *ltoa(long n)
{
        static char bob[14];
        memset(bob,0,14);

        sprintf(bob,"%ld",n);

        return bob;
}

//to prettify things for the interper,
//kill off any \n's.
int strip_whitespace(LPSAFESTRING lpss)
{
        char *p;	

        for(p=lpss->str; *p != 0;p++)
        {
		if(*p == '\n') 
		{
                        if(*(p+1) == 0) { *p = 0; }
                        else
                        {
                                memmove(p,p+1,strlen(p+1)); //move the string back
                                *(p+strlen(p+1)) = 0; //kill the doubled char.
                        }
		}
        }

        return 0;
}

//return a pointer into the string at offset.
char *getstrptr(LPSAFESTRING lpss,int offset)
{
	char *p1;
	p1 = lpss->str;
	if(!p1) return 0;
	p1 += offset;
	return p1;
}

//(re)allocate the string to a certain length....
LPSAFESTRING setlength(LPSAFESTRING lpss,int len)
{
        lpss->str = realloc(lpss->str,len);
	lpss->len = len;
	if (!lpss->str) lpss = 0;

	return lpss;
}


//initialize and wipe string
int zerostring(LPSAFESTRING lpss)
{
        lpss->str = 0;
        setlength(lpss,1);
        memset(lpss->str,0,lpss->len);
	lpss->actual_len = 0;
        return 0;
}



//take out a portion of a safestring and replace it with another str.
int replacestr(LPSAFESTRING lpss,int offset,int length,char *newstr)
{
        SAFESTRING tmp;

        zerostring(&tmp);
        addsafestrn(&tmp,lpss,offset);
        addrawstr(&tmp,newstr);
        addsafestro(&tmp,lpss,offset+length);

        zerostring(lpss);
        addsafestr(lpss,&tmp);
        killstring(&tmp);

        return 0;
}

//cats a raw string to the safestring
int addrawstr(LPSAFESTRING lpss,char *str)
{
	if(strlen(lpss->str)+strlen(str)+1 > lpss->len)
	{
		//need to alloc more space.
		lpss = setlength(lpss,strlen(lpss->str)+strlen(str)+1);
		if(!lpss) return 1;
	}
	strcat(lpss->str,str);
	lpss->actual_len = strlen(lpss->str);
	return 0;
}

//cats a raw string to the safestring up to 'n' chars
int addrawstrn(LPSAFESTRING lpss,char *str,int n)
{
	int mx;
	mx = strlen(str);
	if(n < mx) mx = n;

	if(strlen(lpss->str)+mx+1 > lpss->len)
	{
		//need to alloc more space.
		lpss = setlength(lpss,strlen(lpss->str)+mx+1);
		if(!lpss) return 1;
	}

	strncat(lpss->str,str,n);
	lpss->actual_len = strlen(lpss->str);

	return 0;
}

//cats a safestring to another safestring.
int addsafestr(LPSAFESTRING lptarget,LPSAFESTRING lpsource)
{
	if(strlen(lptarget->str)+strlen(lpsource->str)+1 > lptarget->len)
	{
		//alloc more space
                lptarget = setlength(lptarget,strlen(lptarget->str)+strlen(lpsource->str)+1);
                if(!lptarget) return 1;
	}
	strcat(lptarget->str,lpsource->str);
	lptarget->actual_len = strlen(lptarget->str);
	return 0;
}

//cats a safestring to another safestring from an offset
int addsafestro(LPSAFESTRING lptarget,LPSAFESTRING lpsource,int offset)
{
        if(strlen(lptarget->str)+strlen(lpsource->str+offset)+1 > lptarget->len)
	{
		//alloc more space
                lptarget = setlength(lptarget,strlen(lptarget->str+offset)+strlen(lpsource->str)+1);
                if(!lptarget) return 1;
	}
        strcat(lptarget->str,lpsource->str+offset);
	lptarget->actual_len = strlen(lptarget->str);
	return 0;
}

//cats a safestring to another safestring up to n chars.
int addsafestrn(LPSAFESTRING lptarget,LPSAFESTRING lpsource,int n)
{
	int mx;
	mx = strlen(lpsource->str);
	if(n < mx) mx = n;

	if(strlen(lptarget->str)+mx+1 > lptarget->len)
	{
		//alloc more space
		lptarget = setlength(lptarget,strlen(lptarget->str)+mx+1);
		if(!lptarget) return 1;
	}
	strncat(lptarget->str,lpsource->str,n);
	lptarget->actual_len = strlen(lptarget->str);
	return 0;
}

//cats up to n chars of a safestring to another safestring from an offset
int addsafestrno(LPSAFESTRING lptarget,LPSAFESTRING lpsource,int n,int offset)
{
        int mx;

        mx = strlen(lpsource->str+offset);
        if(n < mx) mx = n;

        if(strlen(lptarget->str)+mx+1 > lptarget->len)
	{
		//alloc more space
                lptarget = setlength(lptarget,strlen(lptarget->str)+mx+1);
                if(!lptarget) return 1;
	}
        strncat(lptarget->str,lpsource->str+offset,n);
	lptarget->actual_len = strlen(lptarget->str);
	return 0;
}



//free it.
int killstring(LPSAFESTRING lpss)
{
	free(lpss->str);
	lpss->str = 0;
	return 0;
}


char *ltrim(char *str)
{
        while(*str)
                if(*str == ' ' || *str == '\n' || *str == '\t') { str++;}
                else { break; }

        return str;                     
}

////////////////////////////////////////////////////////////
// list item stuff.

int kill_entire_list(LPSAFELISTITEM lpFirst)
{
	LPSAFELISTITEM lpNext;
	LPSAFELISTITEM lpThis;

	lpThis = lpFirst;
	while(lpThis)
	{
		lpNext = lpThis->next;
		kill_list_item(lpThis);
		lpThis = lpNext;
	}

	return 0;
}

int kill_list_item(LPSAFELISTITEM lpss)
{
	free(lpss->str);
	free(lpss);
	return 0;
}

LPSAFELISTITEM create_list_item(LPSAFELISTITEM lpPrev)
{
	LPSAFELISTITEM lpCur;

	lpCur = calloc(sizeof(SAFELISTITEM),1);
        lpCur->len = 0;
//        lpCur->str = calloc(1,1);
        lpCur->next = 0;
	lpCur->sub_line_no = 0;

        if(lpPrev) lpPrev->next = lpCur;
        lpCur->prev = lpPrev;

	return lpCur;
}

LPSAFELISTITEM dup_list_item(int line_no, int sub_line_no)
{
	//duplicate the line in the chain with this number-set.
	//if sub_line_no == -1, then use the *last* line in the chain
	//with this number.

	LPSAFELISTITEM pThis = pFirstItem;
	LPSAFELISTITEM pNew;
	LPSAFELISTITEM pFound = 0;
	LPSAFELISTITEM pNext;

	if(sub_line_no == -1)
	{
		while(pThis)
		{
			if(pThis->line_no == line_no) pFound = pThis;
			if(pThis->line_no > line_no) break;
			pThis = pThis->next;
		}
		pThis = pFound;
	}
	else
	{
		while(pThis)
		{
			if(pThis->line_no == line_no && pThis->sub_line_no == sub_line_no)
				break;
			if(pThis->line_no > line_no) 
			{
				pThis = pThis->prev;
				break;
			}
			pThis = pThis->next;
		}
	}

	if(!pThis) return 0;
	
	pNew = calloc(sizeof(SAFELISTITEM),1);
	pNew->len = 0;
	pNew->str = calloc(1,1);
	pNew->type = 2;
	pNew->sub_line_no = pThis->sub_line_no + 1;
	pNew->line_no = pThis->line_no;
	pNew->user_line_no = pThis->user_line_no;

	update_list_item(pNew,pThis->str);

	//insert into chain
	pNew->next = pThis->next;
	pNew->prev = pThis;

	if(pThis->next) pThis->next->prev = pNew;
	pThis->next = pNew;

	if(pLastItem == pThis) pLastItem = pNew;

	if(pNew->next)
	{
		//if any lines are after this line, with the same major line no
		//and a different sub line no, then we hafta fix that.
		pNext = pNew->next;
		while(pNext)
		{
			if(pNext->line_no != line_no) break;
			pNext->sub_line_no++;
			pNext = pNext->next;
		}
	}

	return pNew;
}


int update_list_item(LPSAFELISTITEM lpli,char *buf)
{
	if(strlen(buf) > lpli->len)
	{
                lpli->str = realloc(lpli->str,strlen(buf)+1);
                lpli->len = strlen(buf)+1;
	}

	strcpy(lpli->str,buf);
	return 0;
}

int update_list_item_from_safe_str(LPSAFELISTITEM lpli,LPSAFESTRING pBuf)
{
	if(strlen(pBuf->str) > lpli->len)
	{
                lpli->str = realloc(lpli->str,strlen(pBuf->str)+1);
                lpli->len = strlen(pBuf->str)+1;
	}

	strcpy(lpli->str,pBuf->str);
	return 0;
	
}
