/* userfuncs.c
 * contains the functions that the user
 * can access
 *
 * (c) Copyright 2000 Aaron Kimball
 */

//#define LangFunc(x) int x(char *str,FILE *fpin, FILE *fpout,LPSAFELISTITEM pThisLine,unsigned long *force_jump)

SAFESTRING LoopIncAssign;
int LevelsNested=0;

LangFunc(__user_args_required)
{
	//syntax: argsrequired n;
	//n is precise arg count needed for this function.

	long int ret,n;
	char theline[64];
	LPSAFELISTITEM lpLine;

	//argsrequired() degenerates into: 
	//	if $# ne n
	//		error "Invalid argument count.";
	//		retnow;
	//	endif

	if(!str) 
	{
		SysErr("Argument expected.");
		return 0;
	}

        ret = GetLongArg(str,&n);
        print_eval_error(ret);

	memset(theline,0,64);
	sprintf(theline,"if $# ne %ld",n);
	update_list_item(pThisLine,theline);

	lpLine = dup_list_item(pThisLine->line_no,pThisLine->sub_line_no);
	update_list_item(lpLine,"error \"Invalid argument count.\"");
	
	lpLine = dup_list_item(lpLine->line_no,lpLine->sub_line_no);
	update_list_item(lpLine,"retnow");
	
	lpLine = dup_list_item(lpLine->line_no,lpLine->sub_line_no);
	update_list_item(lpLine,"endif");

	*force_jump = global_cur_line;
	sub_line_goto = pThisLine->sub_line_no;
	return 0;
}

LangFunc(__user_print_err)
{
	//syntax:  error "errstr";
	//causes Emu to SysErr() the error string.

        SAFESTRING safestr;

        zerostring(&safestr);
        if(str)
        {
                switch(GetStringArg(str,&safestr))
                {
                case 1:
                        SysErr("Invalid identifier.");
                        break;
                case 2:
                        SysErr("String expected.");
                        break;
                case 3:
                        SysErr("Identifier not found.");
                        break;
                case 4:
                        SysErr("String expected.");
                        break;
		default:
			SysErr(safestr.str);
			break;
                }
        }
	else
	{
		SysErr("String expected.");
	}

	killstring(&safestr);
	
	return 0;
}

LangFunc(__user_get_user_input)
{
	//syntax: input "prompt";
	//returns: a string containing fgets(fpin);
	//(chomped, of course.)

	SAFESTRING safestr;
	char buf[1024];
	char *p2=0;
	int offset;

	zerostring(&safestr);
	if(str) 
	{
	        switch(GetStringArg(str,&safestr))
        	{
	        case 1:
        	        SysErr("Invalid identifier.");
	                break;
        	case 2:
                	SysErr("String expected.");
        	        break;
	        case 3:
        	        SysErr("Identifier not found.");
	                break;
        	case 4:
                	SysErr("String expected.");
        	        break;
		}
        }

	if(strlen(safestr.str))
	{
		//take anything escaped and deal with it.
                for(p2=safestr.str;*p2;p2++)
                {
                        if(*p2 == '\\')
                        {
                                offset = p2-safestr.str;
                                switch(*(p2+1))
                                {
                                case 'n':
                                        offset = p2-safestr.str;
                                        replacestr(&safestr,p2-safestr.str,2,"\n");
                                        p2 = safestr.str + offset;
                                        break;
                                case '&':
                                case '\\':
                                case '\"':
                                        //just delete the slash
                                        offset = p2-safestr.str;
                                        replacestr(&safestr,p2-safestr.str,1,"");
                                        p2 = safestr.str+offset;
                                        break;
                                case 't':
                                        offset = p2-safestr.str;
                                        replacestr(&safestr,p2-safestr.str,2,"\t");
                                        p2 = safestr.str + offset;
                                        break;
                                default:
                                        SysWarn("Invalid escape character.");
                                        break;
                                } //switch(sequence)
                        } //if there's a \escape
                } //check for escape seqs.
		//print it!
		fprintf(fpout,"%s",safestr.str);
	}

	memset(buf,0,1023);

	if(!backgroundmode) fgets(buf,1023,stdin);

	if(buf[strlen(buf)-1] == '\n' || buf[strlen(buf)-1] == '\r') buf[strlen(buf)-1] = 0;

	SetStrRet(buf);

	return 0;
}

LangFunc(__user_get_one_key)
{
	//syntax: getkey;
	//returns: a string consisting of one character, as returned by getchar();

	char buf[2];
	LPFUNCTIONDEF lpFunc;

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type != 5)
	{
		lpFunc->type = 5;
		lpFunc->val.str_val = malloc(sizeof(SAFESTRING));
	}

	memset(buf,0,2);

	if(!backgroundmode) buf[0] = fgetc(stdin);
	
	zerostring(lpFunc->val.str_val);
	addrawstr(lpFunc->val.str_val,buf);

	return 0;
}

LangFunc(__user_str_to_val)
{
	LPFUNCTIONDEF lpFunc;
	double dret;
	SAFESTRING safestr;
	
/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        zerostring(&safestr);
	addsafestrno(&safestr,&str_func_arg,strlen(str_func_arg.str)-2,1);
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

	dret = atof(safestr.str);

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type == 5)
	{
		killstring(lpFunc->val.str_val);
		free(lpFunc->val.str_val);
	}
	lpFunc->type = 2; //real number.

	lpFunc->val.float_val = dret;

	killstring(&safestr);
	pop_arg_string();

	return 0;
}

LangFunc(__user_val_to_str)
{
	//syntax: tostr val
	//returns: a string-ized version of val.

	LPFUNCTIONDEF lpFunc;
	double val;
	char buf[64];

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(!rettype) ret = 2;
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        val = global_double_return;
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type != 5) { lpFunc->val.str_val = malloc(sizeof(SAFESTRING)); zerostring(lpFunc->val.str_val); }
	lpFunc->type = 5; //string

	memset(buf,0,64);
	sprintf(buf,"%g",val);

	addrawstr(lpFunc->val.str_val,buf);

	pop_arg_string();

	return 0;
}

LangFunc(__user_fsize_func)
{
	//syntax: fsize handle
	//returns: (int) size of file specified by handle, in bytes.

	long int handle;
	int ret;
	LPFUNCTIONDEF lpFunc;
	struct stat file_stats;

	if(safemode) { SysWarn("File access denied."); return 0; }

	ret = GetLongArg(str,&handle);
	print_eval_error(ret);

	fstat(handle,&file_stats);

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type == 5) { killstring(lpFunc->val.str_val); free(lpFunc->val.str_val); }
	lpFunc->type = 3; //integer.
	lpFunc->val.long_val = file_stats.st_size;

	return 0;	
}

LangFunc(__user_fseek_func)
{
	//syntax: fseek handle, pos, modifier
	//modifier      = 0: absolute position
	//		= 1: relative to current position.
	//does: moves the file handle around.
	//returns: nada.

	int handle;
	int pos;
	int modifier;

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

	if(safemode) { SysWarn("File access denied."); return 0; }

        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(!rettype) ret = 2;
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        handle = global_double_return;
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(!rettype) ret = 2;
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        pos = global_double_return;
	
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(!rettype) ret = 2;
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        modifier = global_double_return;

	switch (modifier)
	{
	case 0:
		lseek(handle,pos,SEEK_SET);
		break;
	case 1:
		lseek(handle,pos,SEEK_CUR);
		break;
	default:
		SysWarn("Undefined seek point.");
		break;
	}

	pop_arg_string();

	return 0;
}

LangFunc(__user_str_mid)
{
	//syntax: mid str, off, len
	//return the len characters beginning at offset off of string str.

	LPFUNCTIONDEF lpVar;
	SAFESTRING safestr;
	int n,off;

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        lpVar = GetUserIdentifier("ret");
        if(lpVar->type != 5)
	{
		lpVar->val.str_val = malloc(sizeof(SAFESTRING));
		zerostring(lpVar->val.str_val);
	}
        lpVar->type = 5; //string
	zerostring(lpVar->val.str_val);

	//get the string argument.
        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

	zerostring(&safestr);
	addsafestr(&safestr,&str_func_arg);

	ret = get_arg_by_num(depth,1,0,0,&rettype);
	if(!rettype) ret = 2;
	if(ret)
	{
		print_arg_error(ret);
		pop_arg_string();
		return 1;
	}

	off = global_double_return;

	ret = get_arg_by_num(depth,2,0,0,&rettype);
	if(!rettype) ret = 2;
	if(ret)
	{
		print_arg_error(ret);
		pop_arg_string();
		return 1;
	}

	n = global_double_return;

	if(off + n > strlen(safestr.str)-2)
	{
		n = strlen(safestr.str) - 2 - off;
	}
	addsafestrno(lpVar->val.str_val,&safestr,n,off+1);

        killstring(&safestr);

	pop_arg_string();

	return 0;
}

LangFunc(__user_instr_func)
{
	//syntax: instr str1, str2
	//return 1 if str2 is in str1, or 0 if str2 is not.

	LPFUNCTIONDEF lpVar;
	SAFESTRING safestr1;
	SAFESTRING safestr2;

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        lpVar = GetUserIdentifier("ret");
        if(lpVar->type == 5)
	{
		killstring(lpVar->val.str_val); 
		free(lpVar->val.str_val); 
	}
        lpVar->type = 3; //int

	//get the string argument.
        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

	zerostring(&safestr1);
	addsafestr(&safestr1,&str_func_arg);

        ret = get_arg_by_num(depth,1,0,0,&rettype);
        if(ret)
        {
		killstring(&safestr1);
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }

	zerostring(&safestr2);
	addsafestr(&safestr2,&str_func_arg);

	//the problem is that both of these strings are outer-quoted.
	//therefore, we have to remove those quotes in at least safestr2.
	safestr2.str[strlen(safestr2.str)-1] = 0;
	replacestr(&safestr2,0,1,"");

	ret = strstr(safestr1.str,safestr2.str) != 0 ? 1 : 0;

	lpVar->val.long_val = ret;

        killstring(&safestr1);
        killstring(&safestr2);

	pop_arg_string();

	return 0;
}

LangFunc(__user_str_right)
{
	//syntax: right str, n
	//return the last n characters of str, or the whole string if n > len str.

	LPFUNCTIONDEF lpVar;
	SAFESTRING safestr;
	int n;

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        lpVar = GetUserIdentifier("ret");
        if(lpVar->type != 5)
	{
		lpVar->val.str_val = malloc(sizeof(SAFESTRING));
		zerostring(lpVar->val.str_val);
	}
        lpVar->type = 5; //string
	zerostring(lpVar->val.str_val);

	//get the string argument.
        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

	zerostring(&safestr);
	addsafestr(&safestr,&str_func_arg);

	ret = get_arg_by_num(depth,1,0,0,&rettype);
	if(!rettype) ret = 2;
	if(ret)
	{
		killstring(&safestr);
		print_arg_error(ret);
		pop_arg_string();
		return 1;
	}

	n = global_double_return;

	if(strlen(safestr.str)-2 < n-1)
	{
		addsafestrno(lpVar->val.str_val,&safestr,strlen(safestr.str)-2,1);
	}
	else
	{
		addsafestrno(lpVar->val.str_val,&safestr,n,strlen(safestr.str)-n-1);
	}

        killstring(&safestr);

	pop_arg_string();

	return 0;
}

LangFunc(__user_str_left)
{
	//syntax: left str, n
	//return the first n characters of str, or the whole string if n > len str.

	LPFUNCTIONDEF lpVar;
	SAFESTRING safestr;
	int n;

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        lpVar = GetUserIdentifier("ret");
        if(lpVar->type != 5)
	{
		lpVar->val.str_val = malloc(sizeof(SAFESTRING));
		zerostring(lpVar->val.str_val);
	}
        lpVar->type = 5; //string
	zerostring(lpVar->val.str_val);

	//get the string argument.
        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
	rettype=0;
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

	zerostring(&safestr);
	addsafestr(&safestr,&str_func_arg);
	rettype=0;
	ret = get_arg_by_num(depth,1,0,0,&rettype);
	if(!rettype)
{
 ret = 2;
}
	if(ret)
	{
		killstring(&safestr);
		print_arg_error(ret);
		pop_arg_string();
		return 1;
	}

	n = global_double_return;

	if(n > strlen(safestr.str)-2) n = strlen(safestr.str)-2;
	addsafestrno(lpVar->val.str_val,&safestr,n,1);

        killstring(&safestr);

	pop_arg_string();

	return 0;
}

LangFunc(__user_stop_exec_func)
{
	//syntax: stop
	//ends program execution

	bSeeking = 0;
	exit(0); // let's go.

	return 0;
}

LangFunc(__user_strlen)
{
	//syntax: len string
	//returns: integer, length of the string 'string'

        SAFESTRING safestr;
	LPFUNCTIONDEF lpVar;
	int n=0;

        zerostring(&safestr);

        switch(GetStringArg(str,&safestr))
        {
        case 1:
                SysErr("Invalid identifier.");
		n = 0;
                break;
        case 2:
                SysErr("String expected.");
		n = 0;
                break;
        case 3:
                SysErr("Identifier not found.");
		n = 0;
                break;
        case 4:
                SysErr("String expected.");
		n = 0;
                break;
	default:
		n = safestr.actual_len;
        }

	lpVar = GetUserIdentifier("ret");
	if(lpVar->type == 5) killstring(lpVar->val.str_val);
	lpVar->type = 4; //unsigned integer.

	lpVar->val.ulong_val = n;
	killstring(&safestr);
	
	return 0;
}

LangFunc(__user_sys_exec_func)
{
	//syntax: exec (commands)
	//uses the shell to execute whatever the user damn well pleases.
	//security notice!!!! --- NEVER SET SUID ON EMU
	//if you do so, Emu will then be able to run root commands, by anyone!
	//TODO: Emu should divest itself of suid on startup
	//if Emu is owned by root and suid, it should change itself to 'nobody:nogroup'

	SAFESTRING safestr;

	if(safemode) { SysWarn("Exec access denied."); return 0; }

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	zerostring(&safestr);

	//first things first: all parameters need to be reconciled, if they're variables.
	//the commands are given in print/fprint format.
	if(do_print_work(str,&safestr) == 10) 
	{
		killstring(&safestr);
		return 0;
	}

	//about time: go for it.
	system(safestr.str);

	killstring(&safestr);

	return 0;
}

LangFunc(__user_eof_func)
{
	//syntax: eof filenum;
	//returns: 1/0 for true/false condition of if reached eof.

	long arg;
	int ret,n;
	char buf[2];
	LPFUNCTIONDEF lpFunc;

	ret = GetLongArg(str,&arg);
	print_eval_error(ret);

	n = read(arg,buf,1); //read a single byte
	lseek(arg,-1,SEEK_CUR); //move backwards by one, to counter-act that.
	
	//now lets give the user something to use :)
	n = !n; //flip it around, so a 1 is eof, a 0 is reading continuable.
	
	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type == 5) { killstring(lpFunc->val.str_val); free(lpFunc->val.str_val); }
	lpFunc->type = 3; //integer.
	lpFunc->val.long_val = n;

	return 0;	
}


LangFunc(__user_read_file_func)
{
	//syntax: read filenum, n;
	//reads n bytes/chars from file 'filenum'.
	//returns: string containing that many chars.
	//(or as many as could be had)

	char *p1=0;
	long int handle;
	long int len;
	SAFESTRING safestr;
	LPFUNCTIONDEF lpFunc;
	int ret;
	int offset;
	char *data;

	if(safemode) { SysWarn("File access denied."); return 0; }

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	p1 = strchr(str,',');
	if(!p1) 
	{
bomb_out:
		//no comma; thus, no second argument. (nothing to print?)
		SysErr("Argument expected.");
		return 0;
	}
	if(!(*p1+1)) goto bomb_out; //no data after the comma. die.

	//isolate the first term (a number) and save it as the handle number.
	zerostring(&safestr);
	addrawstrn(&safestr,str,p1-str);
	ret = eval_expression(safestr.str,&handle,&handle,0,3,0);
	print_eval_error(ret);
	killstring(&safestr);
	if(ret) return ret;

	// now isolate the second term.
	ret = eval_expression(p1+1,&len,&len,0,3,0);
	print_eval_error(ret);
	killstring(&safestr);
	if(ret) return ret;

	if(len < 1)
	{
		SysErr("Invalid argument.");
		return 0;
	}
	
	lseek(handle,0,SEEK_CUR); //switch to read mode by seeking.

	//get space to read to.
	data = malloc(len+1);
	if(!data)
	{
		SysErr("Out of memory.");
		return 0;
	}
	memset(data,0,len+1);

	//read it!
	read(handle,data,len);

        //give it to the user.
        lpFunc = GetUserIdentifier("ret");
        if(lpFunc->type != 5) lpFunc->val.str_val = malloc(sizeof(SAFESTRING));
        lpFunc->type = 5;
        zerostring(lpFunc->val.str_val);
        addrawstr(lpFunc->val.str_val,data);

	//okay. line's read. however. we must now ESCAPE it. (quotes, backslashes.)

	p1 = strchr(lpFunc->val.str_val->str,'\\');
	while(p1)
	{
		offset = p1 - lpFunc->val.str_val->str;
		replacestr(lpFunc->val.str_val,p1-(lpFunc->val.str_val->str),1,"\\\\");
		p1 = strchr(lpFunc->val.str_val->str + offset+2,'\"');
	}

	p1 = strchr(lpFunc->val.str_val->str,'\"');
	while(p1)
	{
		offset = p1 - lpFunc->val.str_val->str;
		replacestr(lpFunc->val.str_val,p1-(lpFunc->val.str_val->str),1,"\\\"");
		p1 = strchr(lpFunc->val.str_val->str + offset+2,'\"');
	}

	return 0;
}

LangFunc(__user_append_file_func)
{
	//syntax: append filenum;
	//seeks to the end of the file, so that you can write to the end 
	//and append it.

	long arg;
	int ret;

	if(safemode) { SysWarn("File access denied."); return 0; }

	ret = GetLongArg(str,&arg);
	print_eval_error(ret);

	lseek(arg,0,SEEK_END);

	return 0;
}

LangFunc(__user_erase_file_func)
{
	//syntax: erase filenum;
	//zero-lengths a file, so you can write to it as a blank doc.

	long arg;
	int ret;

	if(safemode) { SysWarn("File access denied."); return 0; }

	ret = GetLongArg(str,&arg);
	print_eval_error(ret);

	ftruncate(arg,0); //set the file length to zero.
	lseek(arg,0,SEEK_SET);

	return 0;
}

LangFunc(__user_read_file_line_func)
{
	//syntax: readline filenum;
	//returns: a string containing a (chomped) line from a file.

	long arg;
	int ret;
	SAFESTRING safestr;
	LPFUNCTIONDEF lpFunc;
	char buf[3];
	char *p1=0;
	int offset;

	if(safemode) { SysWarn("File access denied."); return 0; }

	ret = GetLongArg(str,&arg);
	print_eval_error(ret);

	zerostring(&safestr);
	lseek(arg,0,SEEK_CUR); //seek before read to flush buffers.

	while(1)
	{
		memset(buf,0,3);
		ret = read(arg,buf,1);
		if(ret == 0)
		{
			//feof
			break;
		}
		if(ret == -1)
		{
			//some sort of error.
			SysErr("Error reading from file");
			killstring(&safestr);
			return 0;
		}
		if(buf[0] == '\n') break;
		addrawstrn(&safestr,buf,1);
	}

	//okay. line's read. however. we must now ESCAPE it. (Quotes and backslashes and ampersands.).
	p1 = strchr(safestr.str,'\\');
	while(p1)
	{
		offset = p1 - safestr.str;
		replacestr(&safestr,p1-(safestr.str),1,"\\\\");
		p1 = strchr(safestr.str + offset+2,'\"');
	}

	p1 = strchr(safestr.str,'\"');
	while(p1)
	{
		offset = p1 - safestr.str;
		replacestr(&safestr,p1-(safestr.str),1,"\\\"");
		p1 = strchr(safestr.str + offset+2,'\"');
	}

	//give it to the user.

	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type != 5) lpFunc->val.str_val = malloc(sizeof(SAFESTRING));
	lpFunc->type = 5;
	zerostring(lpFunc->val.str_val);
	addsafestr(lpFunc->val.str_val,&safestr);

	//clean up.	
	killstring(&safestr);

	return 0;
}

LangFunc(__user_print_to_file_func)
{
	//syntax: fprint handle, (varying arguments)
	//works exactly like print, except takes a file handle

	char *p1=0;
	long int handle;
	SAFESTRING safestr;
	int ret;

	if(safemode) { SysWarn("File access denied."); return 0; }

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}


	p1 = strchr(str,',');
	if(!p1) 
	{
bomb_out:
		//no comma; thus, no second argument. (nothing to print?)
		SysErr("Argument expected.");
		return 0;
	}
	if(!(*p1+1)) goto bomb_out; //no data after the comma. die.

	//isolate the first term (a number) and save it as the handle number.
	zerostring(&safestr);
	addrawstrn(&safestr,str,p1-str);
	ret = eval_expression(safestr.str,&handle,&handle,0,3,0);
	print_eval_error(ret);
	killstring(&safestr);
	if(ret) return ret;

	//pass the rest of the string to do_print_work() to give ourselves the string.
	//then print the string to the file handle we're specifying.

	zerostring(&safestr);

	lseek(handle,0,SEEK_CUR); //make sure its flushed and ready to write.

	if(do_print_work(p1+1,&safestr) == 10)
	{
		killstring(&safestr);
		return 0;
	}
	write(handle,safestr.str,strlen(safestr.str));

	killstring(&safestr);

	return 0;
}

LangFunc(__user_open_file)
{
	//syntax: open "filename";
	//returns: an integer handle to the file.

	//open will open the file with read and write capabilities.
	//(if possible. if not, then attempts again read-only.)
	//essentially the same as "r+" mode with fopen(); 
	//writing permitted, file is not zero-lengthed.
	//(use 'erase' command for that)
	//if file does not exist, it is created wtih -rw-r--r-- (0644)

        SAFESTRING safestr;
	LPFUNCTIONDEF lpFunc;
 	int handle;

	if(safemode) { SysWarn("File access denied."); return 0; }

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

        zerostring(&safestr);

        switch(GetStringArg(str,&safestr))
        {
        case 1:
                SysErr("Invalid identifier.");
                break;
        case 2:
                SysErr("String expected.");
                break;
        case 3:
                SysErr("Identifier not found.");
                break;
        case 4:
                SysErr("String expected.");
                break;
        }


	handle = open(safestr.str,O_RDWR|O_CREAT,0644);

	if(handle == -1)
	{
		handle = open(safestr.str,O_RDONLY,0);
	}

	killstring(&safestr);

	if(handle == -1)
	{
		handle = 0;
	}

	//make the returned number available via 'ret';
	lpFunc = GetUserIdentifier("ret");
	if(lpFunc->type == 5) { killstring(lpFunc->val.str_val); free(lpFunc->val.str_val); }
	lpFunc->type = 3; //int
	lpFunc->val.long_val = handle;

	return 0;
}

LangFunc(__user_close_file)
{
	//syntax: close handlenum;
	//handlenum is an integer, as per use with read()/write() C-style.
	//return: nothing.

	long arg;
	int ret;

	if(safemode) { SysWarn("File access denied."); return 0; }

	ret = GetLongArg(str,&arg);
	print_eval_error(ret);

	if(arg < 3)
	{
		//trying to close a std* handle. no no.
		SysErr("Can not close priviliged handle.");
		return 0;
	}

	close(arg);

	return 0;
}

LangFunc(__user_break_loop)
{
	//1: determine if we have any parameters. if so, isolate and turn into variable.
	//	if we have two parms, die.
	//	if parm1 == 0, die.
	//2: set LoopsEncountered = number of levels to break out.
	//3: set bSeeking = 6. 
	//	in this mode, lines of 'goto %break_xx' (or 'for' or 'while') increase LoopsEncountered.
	//		  and lines of 'label %break_xx' (or 'loop') decrease LoopsEncountered.
	//	only 'include' is valued as a function to execute.
	//4: if LoopsEncountered == 0, then we've found the point to break to. set it as force_jump, 
	//	and end bSeeking mode.
	//5: unload any if statements, etc off of the stack. insert dummy if's to fill the stack
	//	up to the label's if-level (a la __user_goto)

	char *p1 = 0;
	char *p2 = 0;
	int breaklvls = 1;
	SAFESTRING safestr;
	LPFUNCTIONDEF pFunc;

	if(str)
	{
		for(p1 = str; *p1; p1++)
		{
			if(*p1 != ' ' && *p1 != '\t' && *p1 != '\n' && *p1) break;
		}
	}

	//step 1. check for arguments.
	if(p1)
	{
		if(strlen(p1))
		{
			//we have args.
			if(isdigit(*p1)) 
			{
				breaklvls = atoi(p1); //its a number
			}	
			else
			{
				//of course. a variable. resolve it.
				zerostring(&safestr);
				for(p2=p1;*p2;p2++)
				{
					if(!isalnum(*p2) && *p2 != '_') break;
				}
				addrawstrn(&safestr,p1,p2-p1);
				pFunc = GetUserIdentifier(safestr.str);
				killstring(&safestr);
				if(!pFunc)
				{
					SysErr("Identifier not found.");
					return 1;
				}
				switch(pFunc->type)
				{
				case 2:
					breaklvls = pFunc->val.float_val;
					break;
				case 3:
					breaklvls = pFunc->val.long_val;
					break;
				case 4:
					breaklvls = pFunc->val.ulong_val;
					break;
				default:
					SysErr("Invalid identifier type.");
					return 1;
				} //switch pFunc->type
			} //if variable argument
		} //if arguments exist
	} //if an arg string exists.

	if(breaklvls < 0) breaklvls = 0;
	if(breaklvls > LevelsNested) breaklvls = LevelsNested;

	if(breaklvls == 0) 
	{
		SysWarn("Break statement has no effect.");
		return 0;
	}

	//step 2.
	LoopsEncountered = breaklvls;

	//step 3.
	bSeeking = 6;

	return 0;
}

LangFunc(__user_UseMod)
{
	//UseModule -- load a dll which contains emu-specific commands.
	//syntax: UseModule "modulename";

	SAFESTRING safestr,str2;
	int res;
	

	zerostring(&safestr);
	
	switch(GetStringArg(str,&safestr))
	{
	case 1:
		SysErr("Invalid identifier.");	
		break;
	case 2:
		SysErr("String expected.");
		break;
	case 3:
		SysErr("Identifier not found.");
		break;
	case 4:
		SysErr("String expected.");
		break;
	}


	if(!strstr(safestr.str,".so") && !(strstr(safestr.str,".dll")) && !(strchr(safestr.str,'\\')))
	{
		//no filename extension. no path markers.
		res = create_dll_entry(safestr.str);
		switch(res)
		{
		case 2:
			//couldn't load the library. fair enuf. 
			//add a filename ending.
#ifdef TARGET_LINUX
			addrawstr(&safestr,".so");
#endif
#ifdef TARGET_WIN32
			addrawstr(&safestr,".dll");
#endif
			//and a libemu* prefix.
			zerostring(&str2);
			addrawstr(&str2,"libemu");
			addsafestr(&str2,&safestr);
			zerostring(&safestr);
			addsafestr(&safestr,&str2);
			killstring(&str2);
			break;
		case 1:
			SysErr("Out of memory.");
			break;
                case 3:
                        SysErr("RegisterLibrary() symbol not found.");
             	        SysErr("Library does not conform to standards.");
			SysErr("Please notify this module's author of this problem.");
			break;
               	default:
			killstring(&safestr);
			return 0;
                              	break; //we're done.
		}
		
	}

	res = create_dll_entry(safestr.str);
	switch(res)
	{
	case 1:
		SysErr("Out of memory.");
		break;
	case 2:
		SysErr("Could not load library.");
		SysErr("This is usually because the library file was not found.");
		break;
	case 3:
		SysErr("RegisterLibrary() symbol not found.");
		SysErr("Library does not conform to standards.");
		SysErr("Please notify this module's author of this problem.");
		break;
	default:
		break; //do nothing if no error.
	}

	killstring(&safestr);

	return 0;
}

LangFunc(__user_for_loop)
{
	//syntax: for initial assignment, test condition expression, increment assignment
	//e.g.: for i = $0, i < 9, i += 1

	//the loop itself is split up into several components.
	//0. the increment assignment expression is saved into the LoopIncAssign safestring.
	//1. the initial assignment is broken off into its own item in the function list
	//2. a label of the name "%for_xx" where xx is the line-number of the for loop is inserted.
	//3. the test condition is then put into its own item in the function list in the form:
	//		if (test_condition) eq 0 
	//			goto "%break_xx" <-- xx not known til after seek.
	//		endif
	//   (the label %break_xx is inserted at the `loop' point.)
	//4. The initial assignment is marked as a line to return to and execute, after a seek.
	//5. A seek is conducted for the `loop' keyword.
	
	//when the loop keyword is found:
	//5a. The test condition line is updated, as we now know the 'xx' number for it.
	//5b. A 'label %cont_xx' is added in, so that continue lines have a place to jump.
	//6. The increment assignment from the 'for' is put into a line at this place.
	//7. A 'goto %for_xx' is added in.
	//8. A 'label %break_xx' is added in.

	char *p1=0,*p2=0;
	char label_expr[32];
	SAFESTRING buf;
	LPSAFELISTITEM pNewEntry;
	
	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	p1 = strchr(str,',');
	if(!p1)
	{
		SysErr("Argument expected");
		return 0;
	}

	//First: Turn the current line into only the assignment.
	memset(pThisLine->str,0,strlen(pThisLine->str));
	strncpy(pThisLine->str,str,p1-str);

	//Second: Insert label.
	memset(label_expr,0,32);
	sprintf(label_expr,"label %%for_%ld",pThisLine->line_no);
	pNewEntry = dup_list_item(pThisLine->line_no,-1);
	update_list_item(pNewEntry,label_expr);

	//Third: Insert test condition.
	zerostring(&buf);
	addrawstr(&buf,"if (");
	p2 = strchr(p1+1,',');
	if(!p2)
	{
		SysErr("Argument expected");
		killstring(&buf);
		return 0;
	}
	addrawstrn(&buf,p1+1,p2-p1-1);
	addrawstr(&buf,") eq 0");
	pNewEntry = dup_list_item(pThisLine->line_no,-1);
	update_list_item_from_safe_str(pNewEntry,&buf);
	killstring(&buf);

	pNewEntry = dup_list_item(pThisLine->line_no,-1);
	update_list_item(pNewEntry,"goto %break_xx");

	pNewEntry = dup_list_item(pThisLine->line_no,-1);
	update_list_item(pNewEntry,"endif");

	//Zero: (Yes, its out of order. deal.)
	//Isolate the increment expression. save it into LoopIncAssign
	p2++;
	while(*p2 == ' ' || *p2 == '\t') p2++;
	if(!(*p2))
	{
		SysErr("Argument expected");
		return 0;
	}
	
	zerostring(&LoopIncAssign);
	addrawstr(&LoopIncAssign,p2);
	
	//Four: Flag the return-from-seek point.
	seekline = pThisLine->line_no;
	userseekline = pThisLine->user_line_no;

	//Five: Switch into seek mode (5)
	bSeeking = 5;
	LoopsEncountered = 0;

	return 0;
}

LangFunc(__user_while_loop)
{
	char label_expr[32];
	char *p1;
	LPSAFELISTITEM pNewEntry;
	SAFESTRING buf;

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	//First: null out the increment expression (There is none)
	zerostring(&LoopIncAssign);

        //Second: Insert label as overwrite of current line.
        memset(label_expr,0,32);
        sprintf(label_expr,"label %%for_%ld",pThisLine->line_no);
        update_list_item(pThisLine,label_expr);
	pThisLine->type = 2; // this is now a special line (Allowed to use illegal chars in label)

        //Third: Insert test condition.
	p1 = str;
	while(*p1) if(*p1 != ' ' && *p1 != '\t') break;
	if(!strlen(p1))
        {
                SysErr("Argument expected");
                return 0;
        }
	zerostring(&buf);
	addrawstr(&buf,"if (");
	addrawstr(&buf,str);
	addrawstr(&buf,") eq 0");
        pNewEntry = dup_list_item(pThisLine->line_no,-1);
        update_list_item_from_safe_str(pNewEntry,&buf);
	killstring(&buf);

        pNewEntry = dup_list_item(pThisLine->line_no,-1);
        update_list_item(pNewEntry,"goto %break_xx");

        pNewEntry = dup_list_item(pThisLine->line_no,-1);
        update_list_item(pNewEntry,"endif");

        //Four: Flag the return-from-seek point.
        seekline = pThisLine->line_no;
        userseekline = pThisLine->user_line_no;

        //Five: Switch into seek mode (5)
        bSeeking = 5;
        LoopsEncountered = 0;
 
	return 0;
}

LangFunc(__user_continue_loop)
{
	//we hit a continue. we're going to degenerate
	//the current line into 'goto %cont_xx'
	char buf[32];

	memset(buf,0,32);
	sprintf(buf,"goto %%cont_%d",seekline);
	update_list_item(pThisLine,buf);
	pThisLine->type = 2; //make it a magic line.

	return 0;
}


LangFunc(__user_loop_ret)
{

	//we've found the 'loop' component to a for/while loop.
	//time to finish up the job, whot's whot.

	LPSAFELISTITEM pNew,pNext;
	char buf[32];

	//step 5a: update the test condition jump line, as we now have the loop end.
	pNext = pFirstItem;
	while(pNext)
	{
		if(pNext->line_no == seekline)
		{
			//might have it. gotta check if its the right subline.
			if(!strncmp(pNext->str,"goto ",5))
			{
				//yep.
				memset(buf,0,32);
				sprintf(buf,"goto %%break_%ld",pThisLine->line_no);
				update_list_item(pNext,buf);
				break;
			}
		}
		pNext = pNext->next;
	}

	//step 5b. Replace the 'loop' line with the %cont_xx line.
	memset(buf,0,32);
	sprintf(buf,"label %%cont_%d",seekline);
	update_list_item(pThisLine,buf);
	pThisLine->type = 2; //make it a magic line.

	//step six. Dup in the increment expression line.
	if(strlen(LoopIncAssign.str))
	{
		pNew = dup_list_item(pThisLine->line_no,-1);
		update_list_item_from_safe_str(pNew,&LoopIncAssign);
	}

	pNew = dup_list_item(pThisLine->line_no,-1);

	//seven. add the 'goto %for_xx' line.
	memset(buf,0,32);
	sprintf(buf,"goto %%for_%d",seekline);
	update_list_item(pNew,buf);

	//eight. add the 'label %break_xx' line.
	pNew = dup_list_item(pThisLine->line_no,-1);
	memset(buf,0,32);
	sprintf(buf,"label %%break_%ld",pThisLine->line_no);
	update_list_item(pNew,buf);

	//finally, schedule a jump back to 'seekline', so we can execute the loop!
	*force_jump = seekline;
	sub_line_goto = 0;

	return 0;
}


LangFunc(__user_goto_func)
{
	//syntax: goto labelname

	//note that the label specified must be at the current calldepth.
	//(you can only go to labels that are in the current function.)

	//the goto keyword must walk the label list, search for the label
	//with a matching hash, check its calldepth, and jump to it. (set force_jump)

	//if the label specified does not exist, then its possible that its 
	//lower down and hasn't been reached yet. in this eventuality, set bSeeking to 4.
	//(respects: func, return, and label)
	// * if we're in a function, then either a func or a return causes a bomb on the label.
	// * if we're not in a function (we're in global/main) then labels within that function
	//   are to be disregarded. 

	char *p1;

	if(!str)
	{
		SysErr("Expected: label");
		return 0;
	}

	if(pThisLine->type == 2)
	{
		//internally generated label. name begins with '%'
		for(p1=str;*p1;p1++) if(*p1 == '%') break;
                if(strncmp(p1,"%for_",5)==0) LevelsNested--;
 	}
	else
	{
		for(p1=str;*p1;p1++)
        	        if(*p1 != ' ' && *p1 != '\n' && *p1 != '\t') break;
	}
	if(!p1)
	{
		SysErr("Label name expected.");
		return 2;
	}
	
	if(find_label(p1,force_jump,pThisLine))
	{
		//if we're here, we didn't find it.
		bSeeking = 4; //look for it lower down in the file.
		seekline = pThisLine->line_no;
		userseekline = pThisLine->user_line_no;
		labelseekhash = strhash(p1);
		labeldepth = CallDepth;
		zerostring(&labelseekname);
		addrawstr(&labelseekname,p1);
	}

	return 0;
}


LangFunc(__user_label_func)
{
	//syntax: label labelname
	int ret;
	char *p1;
        LPLABELLISTITEM pNextLabel;
 
	if(!str)
	{
		SysErr("Expected: label");
		return 0;
	}

	if(pThisLine->type == 2)
	{
		//internally generated label. name begins with '%'
		for(p1=str;*p1;p1++) if(*p1 == '%') break;
		if(strncmp(p1,"%for_",5)==0) LevelsNested++;
		if(strncmp(p1,"%break_",7)==0) LevelsNested--; 
 	}
	else
	{
		for(p1=str;*p1;p1++)
			if(*p1 != ' ' && *p1 != '\n' && *p1 != '\t') break;
	}

	if(!p1)
	{
		ret = 2;
	}
	else
	{
		ret = create_label(p1,pThisLine->line_no,pThisLine->sub_line_no,pThisLine->type);
	}

	switch(ret)
	{
	case 0:
		//it worked.
		if(bSeeking == 4 && strhash(p1)==labelseekhash)
		{
			if(!strcmp(p1,labelseekname.str) && labeldepth == CallDepth)
			{
				//we were searching for a label, and found it.
				//now find the actual label handle. (we need it)
			        pNextLabel = pLastLabel; //search the tree backwards, in order to
                                //allow for multiple recursive calls to this function.
			        while(pNextLabel)
			        {
			                if(pNextLabel->hash == labelseekhash)
			                {
                        			if(strcmp(pNextLabel->name.str,p1))
			                        {
                        			        //not the right label.
			                                pNextLabel=pNextLabel->pPrev;
                        			        continue;
			                        }
                        			if(pNextLabel->depth == CallDepth)
			                        {
                        			        //we've found our label!
			                                //to jump to it, we have to do three things:
                        			        //a) unload the entire stack down to the function call
			                                //b) read the label's ifdepth and insert dummy stack fillers.
                        			        //c) set the next line to exec!

			                                reset_stack_to_level(pNextLabel->if_depth);
                        			        *force_jump = pNextLabel->line;
							sub_line_goto = pNextLabel->subline;
			                                IfDepth = pNextLabel->if_depth;
                        			        CurIfEvaluation = 0;
							bSeeking = 0;

			                                return 0;
			                        }
			                }

			                pNextLabel=pNextLabel->pPrev;
			        }
			}
		}
		break;
	case 1:
		SysErr("Out of memory.");
		break;
	case 2:
		SysErr("Invalid characters in label name.");
		break;
	case 3: 
		SysErr("Duplicate label name found.");
		break;
	}

	return ret;
}

LangFunc(__user_if)
{
	IfDepth++;
	push(&calls,1);
	push(&calls,bSeeking);
	push(&calls,CurIfEvaluation);
	push(&calls,0);
	push(&calls,0);

	//if we're not executing, dont evaluate the expression
	if(bSeeking) return 0;

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}


	CurIfEvaluation = IfDepth;

	//if true, bSeeking = 0	
	//if false, set bSeeking = 3. (looking for an elseif/else)
	switch(EvaluateComparison(str,fpin,fpout,pThisLine))
	{
	case 0:
		bSeeking = 3;
		break;
	case 1:
		bSeeking = 0;
		break;
	case 10:
		return 0; //"function not complete" -- return immediately for reprocessing.
	}

	return 0;
}

LangFunc(__user_elseif)
{
	//if ifdepth = 0, die. (error)
	//if CurIfEvaluation = 1, set bSeeking = 2.
	//if bSeeking, return

	//evaluate.

	//if not true set bSeeking = 3. (Looking for next elseif/else)
	//if true, set CurIfEvaluation = 1, bSeeking = 0

	if(IfDepth == 0)
	{
		SysErr("Unbound elseif.");
		return 1;
	}

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	if(CurIfEvaluation != IfDepth) return 0;

	switch(bSeeking)
	{
	case 0:
		bSeeking = 2;
		return 0;
	case 2:
		return 0;
	case 3:
		bSeeking = EvaluateComparison(str,fpin,fpout,pThisLine) ? 0: 3;
		break;
	}

	return 0;
}

LangFunc(__user_else)
{
	//if ifdepth == 0,  bomb

	//if CurIfEvaluation == 0, execute.	
	//otherwise, bSeeking = 2. (looking for endif)

	//if executing, set CurIfEvaluation = 1;
	if(IfDepth == 0)
	{
		SysErr("Unbound elseif.");
		return 1;
	}


        if(CurIfEvaluation != IfDepth) return 0;

        switch(bSeeking)
        {
        case 0:
                bSeeking = 2;
                return 0;
        case 2:
                return 0;
        case 3:
                bSeeking = 0;
                break;
        }

	return 0;
}

LangFunc(__user_endif)
{
	if(IfDepth > 0) IfDepth--;

	//if the program was seeking (bSeeking == 2 || 3), and not executing, 
	//read off the ifexec stack. if it was exec'ing, turn execution back on.

	//note: be sure to pop off the top of the ifexec stack anyway.

	//pop CurIfEvaluation off of ifeval stack
	pop(&calls); //throw away empty values
	pop(&calls); //(stack push is padded to 20 bytes to equal func call, for alignment)
	CurIfEvaluation = pop(&calls);
	bSeeking = pop(&calls);
	pop(&calls); //throw away if-control-modifier

	return 0;
}

LangFunc(__user_func_creation)
{
        //the user can create his own functions with this..
        //func runs from here til 'end func'
        //until that point, emu should read it all in,
        //but not execute it.

        //syntax:
        //func functionname
        //   ...
        //return

        //ok. this can be three things.
        //1. if we're hitting this for the first time, create the function
        //2. we'll hit this line if the user called this function.
        //3. we'll hit this line on a second run thru this part of the file
        //     (skip it)

        //take the funcname. test it vs. all existing. if one exists as
        //as usermade func, and line_no == pThisLine->line_no, then
        //we're executing.

        //if line_no differs, then they're trying to create two copies
        //of the func. bomb.

        //if the func doesn't already exist, set it to existant,
        //and set bInFunc=1 so we don't actually execute it.

        SAFESTRING safestr;
        LPFUNCTIONDEF lpNextFunc;
        char *p1=0, *p2=0;

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}


        if (bInFunc == 1 || 
            (ExecFunc != -1 && ExecFunc != pThisLine->line_no))
        {
                //we're already inside a function.
                SysErr("Functions can not be nested.");
                return 1;
        }


        for(p1 = str;*p1;p1++)
        {
                if(*p1 == ' ' || *p1 == '\t') continue;
                if(isalpha(*p1) || *p1 == '_') break;

                SysErr("Invalid characters in identifier.");
                return 1;
        }
        if(!(*p1))
        {
                SysErr("Identifier expected.");
                return 1;
        }

        for(p2=p1;*p2;p2++)
        {
                if(!(*p2)) break;
                if(isalnum(*p2) || *p2 == '_') continue;
                if(*p2 == ' ' || *p2 == '\t') break;

                SysErr("Invalid characters in identifier.");
                return 1;
        }

        zerostring(&safestr);
        addrawstrn(&safestr,p1,p2-p1);

        lpNextFunc = GetUserIdentifier(safestr.str);

        if(lpNextFunc == 0)
        {
                //we're creating the function.
                bInFunc = 1;
                lpNextFunc = create_user_function(safestr.str,pThisLine->line_no);
		if(lpNextFunc == 0)
		{
			SysErr("Error: unexpected; (memory?)");
			killstring(&safestr);
			return 0;
		}
        }

        //function exists; check validity
        if(lpNextFunc->ptr.line == pThisLine->line_no)
        {
                if(ExecFunc == -1)
                {
                        //we're hitting this on a multiple run thru.
                        bInFunc = 1;
                        return 0;
                }

                //we're entering the function.
                //move along, move along....
                ExecFunc = lpNextFunc->ptr.line;
                CallDepth++;
        }
        else
        {
                SysErr("Identifier exists.");
                killstring(&safestr);
                return 1;
        }


        killstring(&safestr);
        return 0;
}

LangFunc(__user_endf_func)
{
	long a1,a2,a3,a4,a5;
    	int depth;
        int rettype;
        int ret,i=0;
	LPFUNCTIONDEF lpFunc;
	LPFUNCARGS pNewLastArgs;

        //every function has to end someplace.
        //and 'return' is the logical place to do it.
        //syntax:
        //{return|retnow} [return value to pass back in ret]

        //if we're in an ExecFunc,
        //pop the calling function off the stack, and adjust the next
        //point of execution.
	//also, destroy any local variables (kill all variables in the list with domain==CallDepth)

        //if we're in a bInFunc,
        //we're just fast-forwarding. bInFunc = 0

        if(bInFunc == 1) { bInFunc = 0; }
	else
	{
		if(!str) goto after_args;
		if(strlen(str))
		{

			//we are returning out of a function, so we need to set
			//the value for the global variable 'ret'
	        	depth = add_arg_string(str);
			if(depth < 0) return 0;
			
			//check if there's anything worthwhile.
			get_arg_count(depth,&i);
			if(!i)
			{
				//no arguments.
				pop_arg_string();
				goto after_args;
			}

	        	//get the args
		        ret = get_arg_by_num(depth,0,0,0,&rettype);

	        	if(ret)
	        	{
		                print_arg_error(ret);
                		pop_arg_string();
        	        	return 1;
		        }

			lpFunc = GetUserIdentifier("ret");
			switch(rettype)
			{
			case 0:
				//string.
				if(lpFunc->type != 5) lpFunc->val.str_val = malloc(sizeof(SAFESTRING));
				lpFunc->type = 5; 
				zerostring(lpFunc->val.str_val);
				addsafestr(lpFunc->val.str_val,&str_func_arg);
				break;
			case 1:
				//some sorta number. make it a float.
				if(lpFunc->type == 5) 
				{
					killstring(lpFunc->val.str_val);
					free(lpFunc->val.str_val);
					lpFunc->val.str_val=0;
				}	
				lpFunc->type = 2;
				lpFunc->val.float_val = global_double_return;
				break;
			default:
				//unsupported type (like an array)
				print_arg_error(2);
			}
			pop_arg_string();
		} //assuming there are args to return
	} //we're actually returning from a func.

after_args:

	//assuming we're returning out of a function we need to clean up its stack usage.

        if(ExecFunc != -1)
        {
		a5 = 1;
		while(a5)
		{
			//keep popping until all loops, if-blocks, etc, are cleared.
			a1 = pop(&calls); //ifdepth
			a2 = pop(&calls); //sub line goto
			a3 = pop(&calls); //force_jump
			a4 = pop(&calls); //execfun
			a5 = pop(&calls); //identifier (1 = if-block, 0 = function)
			if(a5 == -2) break;
		}

		//restore all position variables to how things were pre-function call:
		IfDepth = a1;
		sub_line_goto = a2;
                *force_jump = a3;
                ExecFunc = a4;
                if(ExecFunc == -2)
                {
                        SysErr("Too many function returns.");
                        ExecFunc = -1;
                        *force_jump = -2;
                        return 1;
                }

		kill_vars_at_depth(CallDepth);
		kill_labels_at_depth(CallDepth); //invalidate labels in the function.
                CallDepth--; //officially jump up one level
		pNewLastArgs = pLastFuncArg->prev;
		kill_arg_list_item(pLastFuncArg); //and finally, kill the params to the function.
		pLastFuncArg = pNewLastArgs;
        }

        return 0;
}

LangFunc(__user_include_func)
{
        //includes another source file at current point
        //and executes it in-place.

        //syntax:
        //include "filename";

        SAFESTRING safestr;
        FILE *fp;
        LPFILESTATE lpFS;

        zerostring(&safestr);

        switch(GetStringArg(str,&safestr))
        {
        case 1:
                SysErr("Invalid identifier.");
                break;
        case 2:
                SysErr("String expected.");
                break;
        case 3:
                SysErr("Identifier not found.");
                break;
        case 4:
                SysErr("String expected.");
                break;
        }

        //save the current file state.
        pThisFile->fp = fpin;
        pThisFile->bCopying = bCopying;
	pThisFile->line_num = global_cur_user_line;

        lpFS = create_file(pLastFile);
        if(!lpFS) //if no mem.
        {
                killstring(&safestr);
                return 1; 
        }

        fp=fopen(safestr.str,"r");
        if(!fp)
        {
                SysErr("File not found.");
                killstring(&safestr);
                kill_file(lpFS);
                return 1;
        }

        pThisFile->fp = fp;
        pThisFile->bCopying = bCopying = 1;
        if(pThisFile->pPrevState)
                pThisFile->file_num = pThisFile->pPrevState->file_num + 1;
        killstring(&safestr);

	cur_file_num++;
	global_cur_user_line = 0;
	next_user_line_number = 0;
	bInScript = 0;

        parse(fp,fpout); //parse the output.


        //ok, that file's done at this point!
        //return to current state!

	bInScript = 1;
	cur_file_num--;
        kill_file(pThisFile); //kill the old state.

        bCopying = pThisFile->bCopying;
	global_cur_user_line = next_user_line_number = pThisFile->line_num;

        //now, at this point,
        //we need to delete this command from the queue
        //in case the user runs back to it.

        //connect the trace links...
        //and set up flow of exec...
        if(pThisLine->prev)
                pThisLine->prev->next = pThisLine->next;
        if(pThisLine->next)
        {
                pThisLine->next->prev = pThisLine->prev;
                *force_jump = pThisLine->next->line_no;
        }

        //poof!
        kill_list_item(pThisLine);

	*force_jump = -2;
        
        return 0;
}

int print_arg_error(int res)
{
        switch (res)
        {
        case 1:
                SysErr("Invalid argument count");
                break;
        case 2:
                SysErr("Invalid argument type");
                break;
        case 3:
                SysErr("Invalid symbols in argument");
                break;
        }

        return res;
}

int print_eval_error(int res)
{
        switch (res) 
        {
        case 1:
                SysErr("Invalid expression.");
                break;
        case 2:
                SysErr("Invalid identifier.");
                break;
        case 3:
                SysErr("Array subscript out of bounds.");
                break;
        case 4:
                SysErr("Out of memory.");
                break;
        case 5:
                SysErr("Identifier expected. 1");
                break;
        case 6:
                SysErr("Illegal use of variable.");
                break;
        case 7:
                SysErr("Division over zero.");
                break;
        case 8:
                SysErr("Operation not supported for type.");
                break;
        case 9:
                SysErr("Non-numerical symbol in use.");
                break;
        }

        return res;
}


LangFunc(__user_eval_string)
{
        //evaluate an expression made of strings
        //syntax:
        // var [+]= [var [+]] [expression [+]]...

        //use 'str' :
        //kill off a left-side equals.
        //reduce variables to quoted-strings
        //skim thru for +'s and tokenate on them
        //addrawstrn on +'s.
	//when we find the end, lpnextfunc=getuiden(funcname),
	//and set val to ntext

        char *p1=0,*p2=0;
        LPFUNCTIONDEF lpNextFunc;
        SAFESTRING funcname;
        SAFESTRING curline;
        SAFESTRING ntext;

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

        //ok, before we get the result, we need a place to put it!
        //search thru the orig string for the varname, then
        //find the var, then assign
        zerostring(&curline);
        addrawstr(&curline,pThisLine->str);
        p1 = strtok(curline.str," \t="); //put a str-end @ a whitespace
        p2 = strtok(NULL,""); //and then isolate the real end
        zerostring(&funcname);
        addsafestr(&funcname,&curline);

        //start out trying to find a left-side equals.
        p2 = str;
        for( p1 = str; *p1 != 0; p1++)
                if( *p1 == '=' ) { p2 = p1+1; break; }
        
        zerostring(&curline);
        //p2 contains start of expression.
	//p1 is the equal-sign. (if it exists)
	if(*p1)	
	{
	        if(*(p1-1) == '+')
        	{
                	//+= style equation; put the cur varname into it.
	                addsafestr(&curline,&funcname);
        	        addrawstr(&curline,"+");
	        }
	}

	addrawstr(&curline,p2);

	zerostring(&ntext);

	if(do_string_eval_work(&curline,&ntext) == 10)
	{
		killstring(&curline);
		killstring(&ntext);
		killstring(&funcname);
		return 0;
	}

	lpNextFunc = GetUserIdentifier(funcname.str);
	if(!lpNextFunc)
	{
		SysErr("Invalid l-value identifier");
		killstring(&curline);
		killstring(&ntext);
		killstring(&funcname);
		return 1;
	}
	zerostring(lpNextFunc->val.str_val);
	addsafestr(lpNextFunc->val.str_val,&ntext);

	killstring(&funcname);
        killstring(&curline);
        killstring(&ntext);

        return 0;
}

int do_string_eval_work(LPSAFESTRING curline,LPSAFESTRING ntext)
{
        int numslashes=0;
        char *p1=0,*p2=0,*p3=0,*p4=0,*p5=0;
        LPFUNCTIONDEF lpNextFunc;
        SAFESTRING tempstr;
	SAFESTRING finalstr;
	int numquotes=0;
	SAFESTRING res_str;
        int foundstr;
	char tmp[32];
	int n;
	int numparens=0;
	int numbrackets;

        //ok. now curline contains the full expression.
        //translate variables into quoted strings.
        //put res into &ntext.
        zerostring(ntext);
        for(p2=curline->str;*p2;p2++)
        {
                if(*p2 == ' ' || *p2 == '\t') continue; //whitespace
		if(*p2 == ';') return 0;
                if(*p2 == '\"')
                {
                        //we're beginning a quoted string. isolate to end
                        //and copy verbatim
                        p2++;
                        for(p3=p2;*p3;p3++)
                        {
                                if(*p3 == '\"')
                                {
                                        //is this the end?
                                        numslashes=0;
                                        for(p1=p3-1;p1>=p2;p1--)
                                        {
                                                if(*p1 == '\\')
                                                {
                                                        numslashes++;
                                                        continue;
                                                }
                                                break;
                                        } //escaping check
                                        if(!(numslashes%2))
                                        {
                                                //we really did find the end.
                                                addrawstrn(ntext,p2,p3-p2);
						p3++;
                                                foundstr=1;
						break;
                                        }
                                } //found end
                        } //loop thru
                } //found beginning of quoted string
                else
                {
			if(*p2 == '(')
			{
				p4 = p2;
				numparens = 0;
				numquotes = 0;
				for(;*p2;p2++)
				{
					//fastforward to end of parenpair.
					if(*p2 == '\"') 
					{
						for(p5=p2-1;p5 > curline->str;p5--)
							if(*p5 != '\\') break;
						if((p2-p5)%2) numquotes++;
					}
					if(*p2 == '(' && !(numquotes%2)) numparens++;
					if(*p2 == ')' && !(numquotes%2))
					{
						numparens--;
						if(numparens == 0) break;
					}
						
				}
			}

			if(*p2 == ')')
			{
				//found a closer. isolate, compute the sub equation.
				zerostring(&tempstr);
				addrawstrn(&tempstr,p4+1,(p2-(p4+1)));
				zerostring(&res_str);
				do_string_eval_work(&tempstr,&res_str);
				zerostring(&finalstr);
				addrawstr(&finalstr,"\"");
				addsafestr(&finalstr,&res_str);
				addrawstr(&finalstr,"\"");
				replacestr(curline,p4-(curline->str),p2-p4+1,finalstr.str);
				killstring(&finalstr);
				killstring(&res_str);
				killstring(&tempstr);
				p2 = curline->str - 1; 
				zerostring(ntext);
				continue;
			}

                        //we've found a variable; translate it!
			numbrackets = 0;
                        for(p3=p2;*p3;p3++)
                        {
                                //loop thru and isolate to end of term
				if(*p3 == '[') numbrackets++;
				if(*p3 == ']') numbrackets--;
				if(numbrackets) continue;
                                if(isalnum(*p3)) continue;
                                if(*p3 == '_') continue;
                                if(*p3 == '#' || *p3 == '*' ||
                                   *p3 == '.')
                                        continue;
                                break;
                        }
                        zerostring(&tempstr);
                        addrawstrn(&tempstr,p2,p3-p2);
                        lpNextFunc = GetUserIdentifier(tempstr.str);
                        if(!lpNextFunc)
                        {
                                SysErr("Invalid string identifier.");
	                        killstring(&tempstr);
				return 1;
                        }
			switch(lpNextFunc->type)
			{
			case 0:
			case 1: //its a function.
				//split the line and use the returned value in the string expr.
				p2 = tempstr.str;
				n = strlen(tempstr.str);
				addrawstr(&tempstr,p3);
				numparens=0;
				for(p4 = tempstr.str;*p4;p4++)
				{
					if(*p4 == '(') numparens++;
					if(*p4 == ')') numparens--;
				}
				if(tempstr.str[strlen(tempstr.str)-1] == ')' && numparens < 0)
					tempstr.str[strlen(tempstr.str)-1] = 0; //take out that terminal ')'
				p3 = tempstr.str + n;
				InsertFunctionCall(p2,p3,5);	
	
                        	killstring(&tempstr);
				//after we've split the line, abort so this can be re-run.
                                return 10;
				break;
			case 2: //float
				memset(tmp,0,32);
				sprintf(tmp,"%g",lpNextFunc->val.float_val);
				addrawstr(ntext,tmp);
				break;
			case 3: //int
				memset(tmp,0,32);
				sprintf(tmp,"%ld",lpNextFunc->val.long_val);
				addrawstr(ntext,tmp);
				break;
			case 4: //uint
				memset(tmp,0,32);
				sprintf(tmp,"%lu",lpNextFunc->val.long_val);
				addrawstr(ntext,tmp);
				break;
			case 5: //string
                        	addsafestr(ntext,lpNextFunc->val.str_val);
				break; 
			default:
                                SysErr("Type mismatch.");
	                        killstring(&tempstr);
                                return 1;
                        }
                        killstring(&tempstr);
                        foundstr = 1;
                }
                if(foundstr)
                {
                        p2=p3;
                        //search thru for whitespace.
                        while(*p2 == ' ' || *p2 == '\t') p2++;

			//are we done?
			if(*p2 == 0 || *p2 == ';') return 0;

                        //search for a + sign
                        if(*p2 != '+')
                        {
                                SysErr("Invalid character");
                                return 1;
                        }
                        foundstr=0;
                }
        } //full string search.
	return 0;
}


LangFunc(__user_eval_expression)
{
        //evaluate a variable expression

        //syntax:
        // var = expression;

        //use 'str' :
        //kill off a left-side equals.
        //trim off whitespaces
        //call the evaluator with what's left

        //use pThisItem->str to determine lside-target
        //dump evaluation into it.

        char *p1=0,*p2=0,*p3=0;
	double dret;
	unsigned long uret;
	signed long iret;
        SAFESTRING funcname;
        SAFESTRING curline;
        LPFUNCTIONDEF lpNextFunc;
        int res;

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

        //start out trying to find a left-side equals.
        p2 = str;
        for( p1 = str; *p1; p1++)
                if( *p1 == '=' ) { p2 = p1+1; break; }

        //now begin searching with p2, which is right of any equalsign.
        //check for another equalsign, which would invalidate the expression
        for( p1 = p2; *p1; p1++)
        {
                if( *p1 == '=' )
                {
                        SysErr("Invalid compound expression.");
                        return 1;
                }
        }

        //and also check for whitespace..
        //(and naturally skip it)
        for( ; *p2; p2++)
                if( *p2 != ' ' && *p2 != '\t') break;

        if( !strlen(p2) )
        {
                SysErr("No expression.");
                return 1;
        }
	p3 = p2;

        //ok, before we get the result, we need a place to put it!
        //search thru the orig string for the varname, then
        //find the var, then assign
        zerostring(&curline);
        addrawstr(&curline,pThisLine->str);
        p1 = strtok(curline.str," \t="); //put a str-end @ a whitespace
        p2 = strtok(NULL,""); //and then isolate the real end
        zerostring(&funcname);

        addsafestr(&funcname,&curline);

        lpNextFunc = GetUserIdentifier(funcname.str);
        //ok, at this point, lpNextFunc contains the variable
        if(!lpNextFunc)
        {
                SysErr("lvalue Identifier not found.");
                killstring(&funcname);
                killstring(&curline);
                return 1;
        }

        if(lpNextFunc->type > 7)
        {
                //its an array
                SysErr("You can not assign to an array.");
                killstring(&funcname);
                killstring(&curline);
                return 1;
        }
        if(lpNextFunc->type == 5)
        {
                //do something with strings..
                SysErr("Assignment not valid for this type.");
                killstring(&funcname);
                killstring(&curline);
                return 1;
        }
        if(lpNextFunc->type < 2)
        {
                //you can't assign to functions
                SysErr("Assignment not valid for this type.");
                killstring(&funcname);
                killstring(&curline);
                return 1;
        }


        //ok, by now we have an expression.
        //evaluate the damn thing!!!
        res = eval_expression(p3,&uret,&iret,&dret,lpNextFunc->type,0);
	if(res == 10) 
	{
		killstring(&funcname);
		killstring(&curline);
	}
	if(res == 11)
	{
		//okay -- we need to autoconvert the recipient variable
		//to a string, and reprocess like so.
		lpNextFunc->val.str_val = malloc(sizeof(SAFESTRING));
		zerostring(lpNextFunc->val.str_val);
		lpNextFunc->type = 5;
		lpNextFunc->ptr.func = __user_eval_string;
		killstring(&funcname);
		killstring(&curline);
		*force_jump = global_cur_line; //schedule this line for re-execution.	
		sub_line_goto = global_cur_subline;
		return 0;
	}

	switch(lpNextFunc->type)
	{
	case 2:
		//float
		lpNextFunc->val.float_val = dret;
		break;
	case 3:	
		//integer (signed)
		lpNextFunc->val.long_val = iret;
		if(!iret && dret) lpNextFunc->val.long_val = (long)dret;
		break;
	case 4:
		//integer, unsigned
		lpNextFunc->val.ulong_val = uret;
		if(!uret && dret) lpNextFunc->val.ulong_val = (unsigned long)dret;
		break;
	}

        print_eval_error(res);

        killstring(&funcname);
        killstring(&curline);

        return res;
}

LangFunc(__user_print_func)
{

	//syntax: print (Varying arguments);
	//basically prints anything to the stdout.

	SAFESTRING safestr;

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	zerostring(&safestr);

	if(do_print_work(str,&safestr) == 10)
	{
		killstring(&safestr);
		return 10;
	}
	fprintf(fpout,"%s",safestr.str);
        fflush(fpout);

	killstring(&safestr);

	return 0;
}

int do_print_work(char *str,LPSAFESTRING output)
{
	//take the argument and print the damn thing!
        //however, we gotta make sure that we check for multiple
        //args, quotes, variables, etc... yuck.

        //syntax:
        //print "str1",var2,"str3",var4,var5...;

	//if an evaluation requires a function call, then the
	//print to be re-evaluated, then do_print_work() returns 10
	//any function receiving that should clean up and exit cleanly.

        char *p1=0,*p2=0,*p3=0;
	char *p4=0,*p5=0;
        double d_res;
        long int i_res;
        int x;
        int offset;
        int numslashes=0;
	int numparens=0;
	int length=0;
        char *n1=0;
        SAFESTRING safestr;
        LPFUNCTIONDEF lpVar;
	char txtnum[32];
	SAFESTRING varname;

        p1 = str;
        while( *p1)
        {
                //ffwd past whitespace
                for( ;*p1;p1++)
                {
                        if(*p1 != ' ' && *p1 != '\t' && *p1 != '\n') break;
                }
                if(!*p1) return 0;

                //we're at the start of a term
                if(*p1 == '\"')
                {
                        //string literal
                        zerostring(&safestr);
                        //search for terminal quote:
                        p1++;
                        for(p3=p1;*p3;p3++)
                        {
                                if(*p3 == '\"')
                                {
                                        numslashes=0;
                                        for(n1 = p3-1; n1 > p1-1;n1--)
                                        {
                                                if(*n1 == '\\') numslashes++;
                                                else break;
                                        }
                                        if(!(numslashes%2)) break;
                                }
                        }

                        if(!(*p3))
                        {
                                SysErr("Unterminated string literal.");
                                zerostring(&safestr);
                                killstring(&safestr);
                                return 1;
                        }
                        addrawstrn(&safestr,p1,p3-p1);

			for(p4=safestr.str;*p4;p4++)
			{
				//expand the string; any '&variable' identifiers are evaluated.
				if(*p4 == '&')
				{
					//check if its escaped.
					for(p5 = p4-1;p5 > safestr.str; p5--)
						{ if(*p5 != '\\') break; }
					if(!((p4-p5)%2)) continue; //its escaped.

					//hey, we found one! find the end!
					for(p5 = p4+1; *p5; p5++)
					{
						if(isdigit(*p5) && p5 == p4+1) 
						{
							SysErr("Invalid character."); 
							killstring(&safestr);
							return 1;
						}
						if(isalnum(*p5)) continue;
						if(*p5 == '_') continue;
						if(*p5 == '$' && p5 == p4+1) continue;
						if((*p5 == '.' || *p5 == '*' || *p5 == '#') && *(p5-1) == '$') continue;
						break;
					}
					//p4+1 marks beginning; p5-1 marks end of identifier.
					//isolate, translate, replacestr()
					zerostring(&varname);
					addrawstrn(&varname,p4+1,p5-p4-1);
					lpVar = GetUserIdentifier(varname.str);
					killstring(&varname);
					if(!lpVar)
					{
						SysErr("Invalid identifier.");
						killstring(&safestr);
						return 1;
					}
					offset = p4-safestr.str;
					switch(lpVar->type)
					{
					case 2:
						memset(txtnum,0,32);	
						sprintf(txtnum,"%g",lpVar->val.float_val);
						replacestr(&safestr,p4-safestr.str,p5-p4,txtnum);
						length = strlen(txtnum);
						break;
					case 3:
						memset(txtnum,0,32);
						sprintf(txtnum,"%ld",lpVar->val.long_val);
						replacestr(&safestr,p4-safestr.str,p5-p4,txtnum);
						length = strlen(txtnum);
						break;
					case 4:
						memset(txtnum,0,32);
						sprintf(txtnum,"%lu",lpVar->val.ulong_val);
						replacestr(&safestr,p4-safestr.str,p5-p4,txtnum);
						length = strlen(txtnum);
						break;
					case 5:
						replacestr(&safestr,offset,p5-p4,lpVar->val.str_val->str);
						length = lpVar->val.str_val->actual_len;
						for(p4=safestr.str+offset;
							p4 < safestr.str+offset+length;p4++)
						{
							if(*p4 == '&')
							{
								//we need to escape this.
								replacestr(&safestr,p4-safestr.str,1,"\\&");
								p4=safestr.str+offset+1;
								length += 1;
							}
						}
						break;
					default:
						replacestr(&safestr,offset,p5-p4,"(null)");
						length = 6;
						break;
					} //switch lpvar->type
					p4 = safestr.str + offset + length - 1;
				} //if there's a variable inside the string.
			} //var inside string check.
print_string_out:
                        for(p2=safestr.str;*p2;p2++)
                        {
                                //also, turn any escape chars into valid stuff.
                                //that is, \n for now.
                                if(*p2 == '\\')
                                {
                                        offset = p2-safestr.str;
                                        switch(*(p2+1))
                                        {
                                        case 'n':
                                                offset = p2-safestr.str;
                                                replacestr(&safestr,p2-safestr.str,2,"\n");
                                                p2 = safestr.str + offset;
                                                break;
					case '&':
                                        case '\\':
                                        case '\"':
                                                //just delete the slash
                                                offset = p2-safestr.str;
                                                replacestr(&safestr,p2-safestr.str,1,"");
                                                p2 = safestr.str+offset;
                                                break;
					case 't':
						offset = p2-safestr.str;
						replacestr(&safestr,p2-safestr.str,2,"\t");
						p2 = safestr.str + offset;
						break;
                                        default:
                                                SysWarn("Invalid escape character.");
                                                break;
                                        } //switch(sequence)
                                } //if there's a \escape
                        } //check for escape seqs.

			addsafestr(output,&safestr);

                        //find another term.
                        for(p1 = p3 + 1;*p1;p1++)
                        {
                                if(*p1 == ',')
                                {
                                        p1++;
                                        if(!(*p1))
                                        {
                                                SysErr("Parameter expected.");
                                                zerostring(&safestr);
                                                killstring(&safestr);
                                                return 1;
                                        }
                                        break;
                                }
                        }
                        if(!(*p1))
                        {
                                zerostring(&safestr);
                                killstring(&safestr);
                                return 0; //no comma, nothing left.
                        }

                        continue;
                } //if in string literal
                else
                {
                        //get the expression..
                        zerostring(&safestr);
			numparens = 0;
                        for(p2 = p1; *p2; p2++)
			{
				if(*p2 == '(') numparens++;
				if(*p2 == ')') numparens--;
                                if(*p2 == ',' && numparens == 0) break;
			}
                        addrawstrn(&safestr,p1,p2-p1);

                        //it may be a string variable
                        //check that here:
                        lpVar = GetUserIdentifier(safestr.str);
                        if(lpVar)
                        {
                                if(lpVar->type == 5)
                                {
                                        //it certainly is.
                                        p3 = p2-1;
                                        zerostring(&safestr);
                                        addsafestr(&safestr,lpVar->val.str_val);
                                        for(p4=safestr.str;*p4;p4++)
                                        {
                                                if(*p4 == '&')
                                                {
							for(p5=p4-1;p5>=safestr.str;p5--)
								if(*p5 != '\\') break;
							if((p4-p5)%2) 
							{
	                                                        //we need to escape this.
        	                                                replacestr(&safestr,p4-safestr.str,1,"\\&");
                	                                        p4=safestr.str;
	                                                }
						}
                                        }

                                        goto print_string_out;
                                }
                        }


                        //it's math equation stuffage. evaluate.

                        //evaluate as internal float..
                        i_res = 0;
                        d_res = 0;
                        x = eval_expression(safestr.str,&i_res,&i_res,&d_res,7,0);
			if(x == 10) //immediate exit.
			{
				zerostring(output);
				killstring(&safestr);
				return 10;
			}
                        if(x)
                        {
                                print_eval_error(x);
				addrawstr(output,"(error)");
                        }
                        else
                        {
                                //print eeet out.
				memset(txtnum,0,32);
                                if(i_res && !d_res)
					sprintf(txtnum,"%ld",i_res);
                                else
					sprintf(txtnum,"%g",d_res);

				addrawstr(output,txtnum);
                        }

                        p1 = p2;
                        p1++;
                }
        } //while there's string left to work with.

        zerostring(&safestr);
        killstring(&safestr);
        return 0;
}

LangFunc(__user_local_var)
{
	//create a local variable. possibly, init it.

	//syntax:
	//local {[u]int[eger] | str[ing] | real} [array] varname [=value] [, var = val, var = val...];

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	if(strncmp(str,"integer ",8)==0) return create_user_variables(str+8,pThisLine,3,CallDepth);
	if(strncmp(str,"int ",4)==0) return create_user_variables(str+4,pThisLine,3,CallDepth);

	if(strncmp(str,"uinteger ",9)==0) return create_user_variables(str+9,pThisLine,4,CallDepth);
	if(strncmp(str,"uint ",5)==0) return create_user_variables(str+5,pThisLine,4,CallDepth);

	if(strncmp(str,"real ",5)==0) return create_user_variables(str+5,pThisLine,2,CallDepth);

	if(strncmp(str,"str ",4)==0) return create_user_variables(str+4,pThisLine,5,CallDepth);
	if(strncmp(str,"string ",7)==0) return create_user_variables(str+7,pThisLine,5,CallDepth);


	if(strncmp(str,"integer\t",8)==0) return create_user_variables(str+8,pThisLine,3,CallDepth);
	if(strncmp(str,"int\t",4)==0) return create_user_variables(str+4,pThisLine,3,CallDepth);

	if(strncmp(str,"uinteger\t",9)==0) return create_user_variables(str+9,pThisLine,4,CallDepth);
	if(strncmp(str,"uint\t",5)==0) return create_user_variables(str+5,pThisLine,4,CallDepth);

	if(strncmp(str,"real\t",5)==0) return create_user_variables(str+5,pThisLine,2,CallDepth);

	if(strncmp(str,"str\t",4)==0) return create_user_variables(str+4,pThisLine,5,CallDepth);
	if(strncmp(str,"string\t",7)==0) return create_user_variables(str+7,pThisLine,5,CallDepth);

	SysErr("Undefined variable type");
	return 0;
}

LangFunc(__user_const_func)
{
	//create a constant variable. uhm. i hope, for the user's sake, that he inits it :)

	//syntax:
	//local {[u]int[eger] | str[ing] | real} varname [=value] [, var = val, var = val...];

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

	if(strncmp(str,"integer ",8)==0) return create_user_variables(str+8,pThisLine,15,CallDepth);
	if(strncmp(str,"int ",4)==0) return create_user_variables(str+4,pThisLine,15,CallDepth);

	if(strncmp(str,"uinteger ",9)==0) return create_user_variables(str+9,pThisLine,16,CallDepth);
	if(strncmp(str,"uint ",5)==0) return create_user_variables(str+5,pThisLine,16,CallDepth);

	if(strncmp(str,"real ",5)==0) return create_user_variables(str+5,pThisLine,14,CallDepth);

	if(strncmp(str,"str ",4)==0) return create_user_variables(str+4,pThisLine,17,CallDepth);
	if(strncmp(str,"string ",7)==0) return create_user_variables(str+7,pThisLine,17,CallDepth);


	if(strncmp(str,"integer\t",8)==0) return create_user_variables(str+8,pThisLine,15,CallDepth);
	if(strncmp(str,"int\t",4)==0) return create_user_variables(str+4,pThisLine,15,CallDepth);

	if(strncmp(str,"uinteger\t",9)==0) return create_user_variables(str+9,pThisLine,16,CallDepth);
	if(strncmp(str,"uint\t",5)==0) return create_user_variables(str+5,pThisLine,16,CallDepth);

	if(strncmp(str,"real\t",5)==0) return create_user_variables(str+5,pThisLine,14,CallDepth);

	if(strncmp(str,"str\t",4)==0) return create_user_variables(str+4,pThisLine,17,CallDepth);
	if(strncmp(str,"string\t",7)==0) return create_user_variables(str+7,pThisLine,17,CallDepth);

	SysErr("Undefined variable type");
	return 0;
}

LangFunc(__user_int_func)
{
        //create a global integer, and possibly init it.

        //syntax:
        //int[eger] [array] varname [= value] [, var = val, var = val...];

        //time for an interesting note:
        //for a cute hack, variables and functions are retained in the
        //same list.
        //variables have types 2-5, for ints, reals, and strings.
        //when initialized, they are created with their proper type,
        //and their handler is an intrinsic function, which evaluates the
        //expression at hand. Therefore:
        //
        //int foo;
        //foo = 12;
        //
        //will see 'foo' as an intrinsic func, and will pass to it
        //'foo' and '= 12' so it can evaluate it..

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}


        return create_user_variables(str,pThisLine,3,0);
}


LangFunc(__user_str_func)
{
        //create a global string

        //syntax:
        //str[ing] [array] varname [= "value"] [, var = "val"...];

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

        return create_user_variables(str,pThisLine,5,0);
}


LangFunc(__user_uint_func)
{
        //create unsigned ints

        //syntax: just like int, except its uint[eger]

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

        return create_user_variables(str,pThisLine,4,0);
}

LangFunc(__user_float_func)
{
        //create a floating point val
        //syntax: just like int, except its "real"

	if(!str)
	{
		SysErr("Expected: argument");
		return 0;
	}

        return create_user_variables(str,pThisLine,2,0);
}

LangFunc(__user_clrscr)
{
#ifdef TARGET_DOS
        clrscr();
#endif
        return 0;
}

LangFunc(__user_gotoxy)
{
        //gotoxy x, y

        int x,y;

/** note: here begins the generic code to retrieve user arguments **/
        int depth;
        int rettype;
        int ret;

        depth = add_arg_string(str);
	if(depth < 0) return 0;
        //get the args
        ret = get_arg_by_num(depth,0,0,0,&rettype);
        if(!rettype) ret = 2;
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        x = global_double_return;
/** and here it ends. don't forget to pop_arg_string() at the end of**
 ** the function though. that's all you need to do. strings are     **
 ** returned in str_func_arg, a SAFESTRING.                         **/

        ret = get_arg_by_num(depth,1,0,0,&rettype);
        if(!rettype) ret = 2;
        if(ret)
        {
                print_arg_error(ret);
                pop_arg_string();
                return 1;
        }
        y = global_double_return;

#ifdef TARGET_DOS
        gotoxy(x,y);
#endif
        
        pop_arg_string();
        return 0;
}

LangFunc(__user_xyzzy)
{
        printf("Nothing happens.\n");
        return 0;
}

void init_user_func_list(void)
{
        //put all of our keywords into the hash list..
	create_intrinsic_keyword("while",__user_while_loop);
	create_intrinsic_keyword("for",__user_for_loop);
	create_intrinsic_keyword("loop",__user_loop_ret);
	create_intrinsic_keyword("local",__user_local_var);
	create_intrinsic_keyword("const",__user_const_func);
	create_intrinsic_keyword("goto",__user_goto_func);
	create_intrinsic_keyword("label",__user_label_func);
        create_intrinsic_keyword("int",__user_int_func);
        create_intrinsic_keyword("integer",__user_int_func);
        create_intrinsic_keyword("str",__user_str_func);
        create_intrinsic_keyword("string",__user_str_func);
        create_intrinsic_keyword("uint",__user_uint_func);
        create_intrinsic_keyword("uinteger",__user_uint_func);
        create_intrinsic_keyword("real",__user_float_func);
        create_intrinsic_keyword("func",__user_func_creation);
        create_intrinsic_keyword("retnow",__user_endf_func);
        create_intrinsic_keyword("return",__user_endf_func);
	create_intrinsic_keyword("if",__user_if);
	create_intrinsic_keyword("endif",__user_endif);
	create_intrinsic_keyword("elseif",__user_elseif);
	create_intrinsic_keyword("else",__user_else);
	create_intrinsic_keyword("is",0);
	create_intrinsic_keyword("eq",0);
	create_intrinsic_keyword("ne",0);
	create_intrinsic_keyword("gt",0);
	create_intrinsic_keyword("ge",0);
	create_intrinsic_keyword("lt",0);
	create_intrinsic_keyword("le",0);
	create_intrinsic_keyword("and",0);
	create_intrinsic_keyword("or",0);
	create_intrinsic_keyword("break",__user_break_loop);
	create_intrinsic_keyword("continue",__user_continue_loop);
	create_intrinsic_keyword("argsrequired",__user_args_required);

	//now insert the "standard library" functions.
#ifndef TARGET_DOS
	create_intrinsic_function("UseModule",__user_UseMod);
#endif
        create_intrinsic_function("print",__user_print_func);
        create_intrinsic_function("include",__user_include_func);
        create_intrinsic_function("%arg",0);
	create_intrinsic_function("%tmp",0);
        create_intrinsic_function("ret",0);
        create_intrinsic_function("clrscr",__user_clrscr);
        create_intrinsic_function("gotoxy",__user_gotoxy);
        create_intrinsic_function("xyzzy",__user_xyzzy);
	create_intrinsic_function("open",__user_open_file);
	create_intrinsic_function("close",__user_close_file);
	create_intrinsic_function("readline",__user_read_file_line_func);
	create_intrinsic_function("fprint",__user_print_to_file_func);
	create_intrinsic_function("erase",__user_erase_file_func);
	create_intrinsic_function("append",__user_append_file_func);
	create_intrinsic_function("read",__user_read_file_func);
	create_intrinsic_function("eof",__user_eof_func);
	create_intrinsic_function("stop",__user_stop_exec_func);
	create_intrinsic_function("exec",__user_sys_exec_func);
	create_intrinsic_function("len",__user_strlen);
	create_intrinsic_function("left",__user_str_left);
	create_intrinsic_function("right",__user_str_right);
	create_intrinsic_function("mid",__user_str_mid);
	create_intrinsic_function("instr",__user_instr_func);
	create_intrinsic_function("fseek",__user_fseek_func);
	create_intrinsic_function("fsize",__user_fsize_func);
	create_intrinsic_function("tostr",__user_val_to_str);
	create_intrinsic_function("toval",__user_str_to_val);
	create_intrinsic_function("input",__user_get_user_input);
	create_intrinsic_function("getkey",__user_get_one_key);
	create_intrinsic_function("error",__user_print_err);

	//now set up system constants.
	create_user_variables("true = 1",0,14,0);
	create_user_variables("false = 0",0,14,0);
}

